# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np

from ..utils import Logger, memreport

from ..timing import function_timer, Timer

from ..op import Operator

from .atm import available_utils

if available_utils:
    from .atm import (
        atm_absorption_coefficient,
        atm_absorption_coefficient_vec,
        atm_atmospheric_loading,
        atm_atmospheric_loading_vec,
    )

from .atm import AtmSim

from toast.mpi import MPI

import toast.qarray as qa


class OpSimAtmosphere(Operator):
    """Operator which generates atmosphere timestreams.

    All processes collectively generate the atmospheric realization.
    Then each process passes through its local data and observes the
    atmosphere.

    This operator is only compatible with TOD objects that can return
    AZ/EL pointing.

    Args:
        out (str): accumulate data to the cache with name
            <out>_<detector>.  If the named cache objects do not exist,
            then they are created.
        realization (int): if simulating multiple realizations, the
            realization index.
        component (int): the component index to use for this noise
            simulation.
        lmin_center (float): Kolmogorov turbulence dissipation scale
            center.
        lmin_sigma (float): Kolmogorov turbulence dissipation scale
            sigma.
        lmax_center (float): Kolmogorov turbulence injection scale
             center.
        lmax_sigma (float): Kolmogorov turbulence injection scale sigma.
        gain (float): Scaling applied to the simulated TOD.
        zatm (float): atmosphere extent for temperature profile.
        zmax (float): atmosphere extent for water vapor integration.
        xstep (float): size of volume elements in X direction.
        ystep (float): size of volume elements in Y direction.
        zstep (float): size of volume elements in Z direction.
        nelem_sim_max (int): controls the size of the simulation slices.
        z0_center (float):  central value of the water vapor
             distribution.
        z0_sigma (float):  sigma of the water vapor distribution.
        common_flag_name (str):  Cache name of the output common flags.
            If it already exists, it is used.  Otherwise flags
            are read from the tod object and stored in the cache under
            common_flag_name.
        common_flag_mask (byte):  Bitmask to use when flagging data
           based on the common flags.
        flag_name (str):  Cache name of the output detector flags will
            be <flag_name>_<detector>.  If the object exists, it is
            used.  Otherwise flags are read from the tod object.
        flag_mask (byte):  Bitmask to use when flagging data
           based on the detector flags.
        apply_flags (bool):  When True, flagged samples are not
             simulated.
        report_timing (bool):  Print out time taken to initialize,
             simulate and observe
        wind_dist (float):  Maximum wind drift before discarding the
            volume and creating a new one [meters].
        cachedir (str):  Directory to use for loading and saving
            atmosphere realizations.  Set to None to disable caching.
        freq (float):  Observing frequency in GHz.
        write_debug (bool):  If True, write debugging files.
    """

    def __init__(
        self,
        out="atm",
        realization=0,
        component=123456,
        lmin_center=0.01,
        lmin_sigma=0.001,
        lmax_center=10,
        lmax_sigma=10,
        zatm=40000.0,
        zmax=2000.0,
        xstep=100.0,
        ystep=100.0,
        zstep=100.0,
        nelem_sim_max=10000,
        gain=1,
        z0_center=2000,
        z0_sigma=0,
        apply_flags=False,
        common_flag_name=None,
        common_flag_mask=255,
        flag_name=None,
        flag_mask=255,
        report_timing=False,
        wind_dist=10000,
        cachedir=None,
        freq=None,
        plot=False,
        write_debug=False,
    ):
        # Call the parent class constructor
        super().__init__()

        self._out = out
        self._realization = realization
        self._component = component
        self._lmin_center = lmin_center
        self._lmin_sigma = lmin_sigma
        self._lmax_center = lmax_center
        self._lmax_sigma = lmax_sigma
        self._gain = gain
        self._zatm = zatm
        self._zmax = zmax
        self._xstep = xstep
        self._ystep = ystep
        self._zstep = zstep
        self._nelem_sim_max = nelem_sim_max
        self._cachedir = cachedir
        self._freq = freq

        self._z0_center = z0_center
        self._z0_sigma = z0_sigma

        self._apply_flags = apply_flags
        self._common_flag_name = common_flag_name
        self._common_flag_mask = common_flag_mask
        self._flag_name = flag_name
        self._flag_mask = flag_mask
        self._report_timing = report_timing
        self._wind_dist = wind_dist
        self._wind_time = None
        self._write_debug = write_debug
        self._plot = plot

    @function_timer
    def exec(self, data):
        """Generate atmosphere timestreams.

        This iterates over all observations and detectors and generates
        the atmosphere timestreams.

        Args:
            data (toast.Data): The distributed data.

        Returns:
            None

        """
        log = Logger.get()
        group = data.comm.group
        for obs in data.obs:
            try:
                obsname = obs["name"]
            except Exception:
                obsname = "observation"
            prefix = "{} : {} : ".format(group, obsname)
            tod = self._get_from_obs("tod", obs)
            comm = tod.mpicomm
            rank = 0
            if comm is not None:
                rank = comm.rank
            site = self._get_from_obs("site_id", obs)
            weather = self._get_from_obs("weather", obs)

            # Get the observation time span and initialize the weather
            # object if one is provided.
            times = tod.local_times()
            tmin = times[0]
            tmax = times[-1]
            tmin_tot = tmin
            tmax_tot = tmax
            if comm is not None:
                tmin_tot = comm.allreduce(tmin, op=MPI.MIN)
                tmax_tot = comm.allreduce(tmax, op=MPI.MAX)
            tmin_tot = np.floor(tmin_tot)
            tmax_tot = np.ceil(tmax_tot)
            weather.set(site, self._realization, tmin_tot)

            key1, key2, counter1, counter2 = self._get_rng_keys(obs)

            absorption = self._get_absorption_and_loading(obs)

            cachedir = self._get_cache_dir(obs, comm)

            if comm is not None:
                comm.Barrier()
            if rank == 0:
                log.debug("{}Setting up atmosphere simulation".format(prefix))

            # Cache the output common flags
            common_ref = tod.local_common_flags(self._common_flag_name)

            scan_range = self._get_scan_range(obs, comm, prefix)

            # Loop over the time span in "wind_time"-sized chunks.
            # wind_time is intended to reflect the correlation length
            # in the atmospheric noise.

            tmr = Timer()
            if self._report_timing:
                if comm is not None:
                    comm.Barrier()
                tmr.start()

            tmin = tmin_tot
            istart = 0
            counter1start = counter1
            while tmin < tmax_tot:
                if comm is not None:
                    comm.Barrier()
                if rank == 0:
                    log.debug(
                        "{}Instantiating atmosphere for t = {}".format(
                            prefix, tmin - tmin_tot
                        )
                    )

                istart, istop, tmax = self._get_time_range(
                    tmin, istart, times, tmax_tot, common_ref, tod, weather
                )

                ind = slice(istart, istop)
                nind = istop - istart

                rmin = 0
                rmax = 100
                scale = 10
                counter2start = counter2
                counter1 = counter1start
                xstart, ystart, zstart = self._xstep, self._ystep, self._zstep

                while rmax < 100000:
                    sim, counter2 = self._simulate_atmosphere(
                        weather,
                        scan_range,
                        tmin,
                        tmax,
                        comm,
                        key1,
                        key2,
                        counter1,
                        counter2start,
                        cachedir,
                        prefix,
                        tmin_tot,
                        tmax_tot,
                        rmin,
                        rmax,
                    )

                    if self._plot:
                        self._plot_snapshots(
                            sim,
                            prefix,
                            obsname,
                            scan_range,
                            tmin,
                            tmax,
                            comm,
                            rmin,
                            rmax,
                        )

                    self._observe_atmosphere(
                        sim,
                        tod,
                        comm,
                        prefix,
                        common_ref,
                        istart,
                        nind,
                        ind,
                        scan_range,
                        times,
                        absorption,
                    )

                    del sim

                    rmin = rmax
                    rmax *= scale
                    self._xstep *= np.sqrt(scale)
                    self._ystep *= np.sqrt(scale)
                    self._zstep *= np.sqrt(scale)
                    counter1 += 1

                if self._write_debug:
                    self._save_tod(
                        obsname, tod, times, istart, nind, ind, comm, common_ref
                    )

                self._xstep, self._ystep, self._zstep = xstart, ystart, zstart
                tmin = tmax

        if self._report_timing:
            if comm is not None:
                comm.Barrier()
            if rank == 0:
                tmr.stop()
                log.debug(
                    "{}Simulate and observe atmosphere:  {} seconds".format(
                        prefix, tmr.seconds()
                    )
                )
        return

    @function_timer
    def _save_tod(self, obsname, tod, times, istart, nind, ind, comm, common_ref):
        import pickle

        rank = 0
        if comm is not None:
            rank = comm.rank

        t = times[ind]
        tmin, tmax = t[0], t[-1]
        outdir = "snapshots"
        if rank == 0:
            try:
                os.makedirs(outdir)
            except FileExistsError:
                pass

        try:
            good = common_ref[ind] & tod.UNSTABLE == 0
        except:
            good = slice(0, nind)

        for det in tod.local_dets:
            # Cache the output signal
            cachename = "{}_{}".format(self._out, det)
            ref = tod.cache.reference(cachename)[ind]
            try:
                # Some TOD classes provide a shortcut to Az/El
                az, el = tod.read_azel(detector=det, local_start=istart, n=nind)
            except Exception as e:
                azelquat = tod.read_pntg(
                    detector=det, local_start=istart, n=nind, azel=True
                )
                # Convert Az/El quaternion of the detector back into
                # angles for the simulation.
                theta, phi = qa.to_position(azelquat)
                # Azimuth is measured in the opposite direction
                # than longitude
                az = 2 * np.pi - phi
                el = np.pi / 2 - theta

            fn = os.path.join(
                outdir,
                "atm_tod_{}_{}_t_{}_{}.pck".format(obsname, det, int(tmin), int(tmax)),
            )
            with open(fn, "wb") as fout:
                pickle.dump([det, t[good], az[good], el[good], ref[good]], fout)

        return

    @function_timer
    def _plot_snapshots(
        self, sim, prefix, obsname, scan_range, tmin, tmax, comm, rmin, rmax
    ):
        """Create snapshots of the atmosphere"""
        from ..vis import set_backend

        set_backend()
        import matplotlib.pyplot as plt
        import pickle

        azmin, azmax, elmin, elmax = scan_range

        # elstep = np.radians(0.01)
        elstep = (elmax - elmin) / 320
        azstep = elstep * np.cos(0.5 * (elmin + elmax))
        azgrid = np.linspace(azmin, azmax, (azmax - azmin) // azstep + 1)
        elgrid = np.linspace(elmin, elmax, (elmax - elmin) // elstep + 1)
        AZ, EL = np.meshgrid(azgrid, elgrid)
        nn = AZ.size
        az = AZ.ravel()
        el = EL.ravel()
        atmdata = np.zeros(nn, dtype=np.float64)
        atmtimes = np.zeros(nn, dtype=np.float64)

        rank = 0
        ntask = 1
        if comm is not None:
            rank = comm.rank
            ntask = comm.size

        r = 0
        t = 0
        my_snapshots = []
        vmin = 1e30
        vmax = -1e30
        tstep = 1
        for i, t in enumerate(np.arange(tmin, tmax, tstep)):
            if i % ntask != rank:
                continue
            err = sim.observe(atmtimes + t, az, el, atmdata, r)
            if err != 0:
                raise RuntimeError(prefix + "Observation failed")
            if self._gain:
                atmdata *= self._gain
            vmin = min(vmin, np.amin(atmdata))
            vmax = max(vmax, np.amax(atmdata))
            atmdata2d = atmdata.reshape(AZ.shape)
            my_snapshots.append((t, r, atmdata2d.copy()))

        outdir = "snapshots"
        if rank == 0:
            try:
                os.makedirs(outdir)
            except FileExistsError:
                pass
        fn = os.path.join(
            outdir,
            "atm_{}_{}_t_{}_{}_r_{}_{}.pck".format(
                obsname, rank, int(tmin), int(tmax), int(rmin), int(rmax)
            ),
        )
        with open(fn, "wb") as fout:
            pickle.dump([azgrid, elgrid, my_snapshots], fout)

        log.debug("Snapshots saved in {}".format(fn), flush=True)

        """
        vmin = comm.allreduce(vmin, op=MPI.MIN)
        vmax = comm.allreduce(vmax, op=MPI.MAX)

        for t, r, atmdata2d in my_snapshots:
            plt.figure(figsize=[12, 4])
            plt.imshow(
                atmdata2d,
                interpolation="nearest",
                origin="lower",
                extent=np.degrees(
                    [0, (azmax - azmin) * np.cos(0.5 * (elmin + elmax)), elmin, elmax]
                ),
                cmap=plt.get_cmap("Blues"),
                vmin=vmin,
                vmax=vmax,
            )
            plt.colorbar()
            ax = plt.gca()
            ax.set_title("t = {:15.1f} s, r = {:15.1f} m".format(t, r))
            ax.set_xlabel("az [deg]")
            ax.set_ylabel("el [deg]")
            ax.set_yticks(np.degrees([elmin, elmax]))
            plt.savefig("atm_{}_t_{:04}_r_{:04}.png".format(obsname, int(t), int(r)))
            plt.close()
        """

        del my_snapshots

        return

    def _get_from_obs(self, name, obs):
        """Extract value for name from observation.

        If name is not defined in observation, raise an exception.

        """
        if name in obs:
            return obs[name]
        else:
            raise RuntimeError(
                "Error simulating atmosphere: observation "
                'does not define "{}"'.format(name)
            )

    def _get_rng_keys(self, obs):
        """
        The random number generator accepts a key and a counter,
        each made of two 64bit integers.
        Following tod_math.py we set
        key1 = realization * 2^32 + telescope * 2^16 + component
        key2 = obsindx * 2^32
        counter1 = hierarchical cone counter
        counter2 = sample in stream (incremented internally in the atm code)
        """
        telescope = self._get_from_obs("telescope_id", obs)
        site = self._get_from_obs("site_id", obs)
        obsindx = self._get_from_obs("id", obs)
        key1 = self._realization * 2 ** 32 + telescope * 2 ** 16 + self._component
        key2 = site * 2 ** 16 + obsindx
        counter1 = 0
        counter2 = 0
        return key1, key2, counter1, counter2

    @function_timer
    def _get_absorption_and_loading(self, obs):
        altitude = self._get_from_obs("altitude", obs)
        weather = self._get_from_obs("weather", obs)
        tod = self._get_from_obs("tod", obs)
        if self._freq is not None:
            if not available_utils:
                msg = (
                    "TOAST not compiled with libaatm support- absorption and "
                    "loading unavailable"
                )
                raise RuntimeError(msg)
            absorption = atm_absorption_coefficient(
                altitude,
                weather.air_temperature,
                weather.surface_pressure,
                weather.pwv,
                self._freq,
            )
            loading = atm_atmospheric_loading(
                altitude,
                weather.air_temperature,
                weather.surface_pressure,
                weather.pwv,
                self._freq,
            )
            tod.meta["loading"] = loading
        else:
            absorption = None

        return absorption

    def _get_cache_dir(self, obs, comm):
        obsindx = self._get_from_obs("id", obs)
        if self._cachedir is None:
            cachedir = None
        else:
            # The number of atmospheric realizations can be large.  Use
            # sub-directories under cachedir.
            subdir = str(int((obsindx % 1000) // 100))
            subsubdir = str(int((obsindx % 100) // 10))
            subsubsubdir = str(obsindx % 10)
            cachedir = os.path.join(self._cachedir, subdir, subsubdir, subsubsubdir)
            if (comm is None) or (comm.rank == 0):
                # Handle a rare race condition when two process groups
                # are creating the cache directories at the same time
                while True:
                    try:
                        os.makedirs(cachedir, exist_ok=True)
                    except OSError:
                        continue
                    except FileNotFoundError:
                        continue
                    else:
                        break
        return cachedir

    @function_timer
    def _get_scan_range(self, obs, comm, prefix):
        tod = self._get_from_obs("tod", obs)
        fp_radius = np.radians(self._get_from_obs("fpradius", obs))
        # Read the extent of the AZ/EL boresight pointing, and use that
        # to compute the range of angles needed for simulating the slab.

        (min_az_bore, max_az_bore, min_el_bore, max_el_bore) = tod.scan_range

        # Use a fixed focal plane radius so that changing the actual
        # set of detectors will not affect the simulated atmosphere.

        elfac = 1 / np.cos(max_el_bore + fp_radius)
        azmin = min_az_bore - fp_radius * elfac
        azmax = max_az_bore + fp_radius * elfac
        if azmin < -2 * np.pi:
            azmin += 2 * np.pi
            azmax += 2 * np.pi
        elif azmax > 2 * np.pi:
            azmin -= 2 * np.pi
            azmax -= 2 * np.pi
        elmin = min_el_bore - fp_radius
        elmax = max_el_bore + fp_radius

        if comm is not None:
            azmin = comm.allreduce(azmin, op=MPI.MIN)
            azmax = comm.allreduce(azmax, op=MPI.MAX)
            elmin = comm.allreduce(elmin, op=MPI.MIN)
            elmax = comm.allreduce(elmax, op=MPI.MAX)

        if elmin < 0 or elmax > np.pi / 2:
            raise RuntimeError(
                "{}Error in CES elevation: elmin = {:.3f} deg, elmax = {:.3f} deg, "
                "elmin_bore = {:.3f} deg, elmax_bore = {:.3f} deg, "
                "fp_radius = {:.3f} deg".format(
                    prefix,
                    np.degrees(elmin),
                    np.degrees(elmax),
                    np.degrees(min_el_bore),
                    np.degrees(max_el_bore),
                    np.degrees(fp_radius),
                )
            )

        return azmin, azmax, elmin, elmax

    @function_timer
    def _get_time_range(self, tmin, istart, times, tmax_tot, common_ref, tod, weather):
        while times[istart] < tmin:
            istart += 1

        # Translate the wind speed to time span of a correlated interval
        wx = weather.west_wind
        wy = weather.south_wind
        w = np.sqrt(wx ** 2 + wy ** 2)
        self._wind_time = self._wind_dist / w

        tmax = tmin + self._wind_time
        if tmax < tmax_tot:
            # Extend the scan to the next turnaround
            istop = istart
            while istop < times.size and times[istop] < tmax:
                istop += 1
            while istop < times.size and (common_ref[istop] | tod.TURNAROUND == 0):
                istop += 1
            if istop < times.size:
                tmax = times[istop]
            else:
                tmax = tmax_tot
        else:
            tmax = tmax_tot
            istop = times.size

        return istart, istop, np.ceil(tmax)

    @function_timer
    def _simulate_atmosphere(
        self,
        weather,
        scan_range,
        tmin,
        tmax,
        comm,
        key1,
        key2,
        counter1,
        counter2,
        cachedir,
        prefix,
        tmin_tot,
        tmax_tot,
        rmin,
        rmax,
    ):
        log = Logger.get()
        rank = 0
        if comm is not None:
            rank = comm.rank
        tmr = Timer()
        if self._report_timing:
            if comm is not None:
                comm.Barrier()
            tmr.start()

        T0_center = weather.air_temperature
        wx = weather.west_wind
        wy = weather.south_wind
        w_center = np.sqrt(wx ** 2 + wy ** 2)
        wdir_center = np.arctan2(wy, wx)

        azmin, azmax, elmin, elmax = scan_range

        sim = AtmSim(
            azmin,
            azmax,
            elmin,
            elmax,
            tmin,
            tmax,
            self._lmin_center,
            self._lmin_sigma,
            self._lmax_center,
            self._lmax_sigma,
            w_center,
            0,
            wdir_center,
            0,
            self._z0_center,
            self._z0_sigma,
            T0_center,
            0,
            self._zatm,
            self._zmax,
            self._xstep,
            self._ystep,
            self._zstep,
            self._nelem_sim_max,
            comm,
            key1,
            key2,
            counter1,
            counter2,
            cachedir,
            rmin,
            rmax,
            write_debug=self._write_debug,
        )

        if self._report_timing:
            if comm is not None:
                comm.Barrier()
            if rank == 0:
                tmr.stop()
                log.debug(
                    "{}OpSimAtmosphere: Initialize atmosphere: {} seconds".format(
                        prefix, tmr.seconds()
                    )
                )
                tmr.clear()
                tmr.start()

        # Check if the cache already exists.

        use_cache = False
        have_cache = False
        if rank == 0:
            if cachedir is not None:
                # We are saving to cache
                use_cache = True
            fname = None
            if cachedir is not None:
                fname = os.path.join(
                    cachedir, "{}_{}_{}_{}.h5".format(key1, key2, counter1, counter2)
                )
            if (fname is not None) and os.path.isfile(fname):
                log.debug(
                    "{}Loading the atmosphere for t = {} from {}".format(
                        prefix, tmin - tmin_tot, fname
                    )
                )
                have_cache = True
            else:
                log.debug(
                    "{}Simulating the atmosphere for t = {}".format(
                        prefix, tmin - tmin_tot
                    )
                )
        if comm is not None:
            use_cache = comm.bcast(use_cache, root=0)

        err = sim.simulate(use_cache=use_cache)
        if err != 0:
            raise RuntimeError(prefix + "Simulation failed.")

        # Advance the sample counter in case wind_time broke the
        # observation in parts

        counter2 += 100000000

        if self._report_timing:
            if comm is not None:
                comm.Barrier()
            if rank == 0:
                op = None
                if have_cache:
                    op = "Loaded"
                else:
                    op = "Simulated"
                tmr.stop()
                log.debug(
                    "{}OpSimAtmosphere: {} atmosphere: {} seconds".format(
                        prefix, op, tmr.seconds()
                    )
                )
                tmr.clear()
                tmr.start()

        return sim, counter2

    @function_timer
    def _observe_atmosphere(
        self,
        sim,
        tod,
        comm,
        prefix,
        common_ref,
        istart,
        nind,
        ind,
        scan_range,
        times,
        absorption,
    ):
        log = Logger.get()
        rank = 0
        if comm is not None:
            rank = comm.rank
        tmr = Timer()
        if self._report_timing:
            if comm is not None:
                comm.Barrier()
            tmr.start()

        azmin, azmax, elmin, elmax = scan_range

        nsamp = tod.local_samples[1]

        if rank == 0:
            log.debug("{}Observing the atmosphere".format(prefix))

        ngood_tot = 0
        nbad_tot = 0

        for det in tod.local_dets:
            # Cache the output signal
            cachename = "{}_{}".format(self._out, det)
            if tod.cache.exists(cachename):
                ref = tod.cache.reference(cachename)
            else:
                ref = tod.cache.create(cachename, np.float64, (nsamp,))

            # Cache the output flags
            flag_ref = tod.local_flags(det, self._flag_name)

            if self._apply_flags:
                good = np.logical_and(
                    common_ref[ind] & self._common_flag_mask == 0,
                    flag_ref[ind] & self._flag_mask == 0,
                )
                ngood = np.sum(good)
            else:
                try:
                    good = common_ref[ind] & tod.UNSTABLE == 0
                    ngood = np.sum(good)
                except:
                    good = slice(0, nind)
                    ngood = nind
            if ngood == 0:
                continue

            try:
                # Some TOD classes provide a shortcut to Az/El
                az, el = tod.read_azel(detector=det, local_start=istart, n=nind)
                az = az[good]
                el = el[good]
            except Exception as e:
                azelquat = tod.read_pntg(
                    detector=det, local_start=istart, n=nind, azel=True
                )[good]
                # Convert Az/El quaternion of the detector back into
                # angles for the simulation.
                theta, phi = qa.to_position(azelquat)
                # Azimuth is measured in the opposite direction
                # than longitude
                az = 2 * np.pi - phi
                el = np.pi / 2 - theta

            if np.ptp(az) < np.pi:
                azmin_det = np.amin(az)
                azmax_det = np.amax(az)
            else:
                # Scanning across the zero azimuth.
                azmin_det = np.amin(az[az > np.pi]) - 2 * np.pi
                azmax_det = np.amax(az[az < np.pi])
            elmin_det = np.amin(el)
            elmax_det = np.amax(el)
            if (
                not (azmin <= azmin_det and azmax_det <= azmax)
                and not (
                    azmin <= azmin_det - 2 * np.pi and azmax_det - 2 * np.pi <= azmax
                )
            ) or not (elmin <= elmin_det and elmin_det <= elmax):
                # DEBUG begin
                import pickle

                with open("bad_quats_{}_{}.pck".format(rank, det), "wb") as fout:
                    pickle.dump(
                        [scan_range, az, el, azelquat, tod._boresight_azel], fout
                    )
                # DEBUG end
                raise RuntimeError(
                    prefix + "Detector Az/El: [{:.5f}, {:.5f}], "
                    "[{:.5f}, {:.5f}] is not contained in "
                    "[{:.5f}, {:.5f}], [{:.5f} {:.5f}]"
                    "".format(
                        azmin_det,
                        azmax_det,
                        elmin_det,
                        elmax_det,
                        azmin,
                        azmax,
                        elmin,
                        elmax,
                    )
                )

            # Integrate detector signal

            atmdata = np.zeros(ngood, dtype=np.float64)

            err = sim.observe(times[ind][good], az, el, atmdata, -1.0)
            if err != 0:
                # Observing failed
                bad = np.abs(atmdata) < 1e-30
                nbad = np.sum(bad)
                log.error(
                    "{}OpSimAtmosphere: Observing FAILED for {} ({:.2f} %) samples. "
                    "det = {}, rank = {}".format(
                        prefix, nbad, nbad * 100 / ngood, det, rank
                    )
                )
                atmdata[bad] = 0
                flag_ref[ind][good][bad] = 255
                nbad_tot += nbad
            ngood_tot += ngood

            if self._gain:
                atmdata *= self._gain

            if absorption is not None:
                # Apply the frequency-dependent absorption-coefficient
                atmdata *= absorption

            ref[ind][good] += atmdata

            del ref

        if comm is not None:
            comm.Barrier()
            ngood_tot = comm.reduce(ngood_tot)
            nbad_tot = comm.reduce(nbad_tot)
        if rank == 0 and nbad_tot > 0:
            log.error(
                "{}: Observe atmosphere FAILED on {:.2f}% of samples".format(
                    prefix, nbad_tot * 100 / ngood_tot
                )
            )
        if self._report_timing:
            if rank == 0:
                tmr.stop()
                log.debug(
                    "{}OpSimAtmosphere: Observe atmosphere: {} seconds".format(
                        prefix, tmr.seconds()
                    )
                )
        return



#!/usr/bin/env python3

import os
import sys
import re
import subprocess as sp
import shutil
import stat

lesson_dir = os.path.dirname(sys.argv[0])
print("Scanning {}".format(lesson_dir))

nersc_dir = "toast-tutorial"

# After running this script, do:
#   rsync -a -e ssh ./toast-tutorial/ dtn01.nersc.gov:/project/projectdirs/cmb/www/toast-tutorial/

for root, dirs, files in os.walk(lesson_dir):
    for d in dirs:
        if re.match(r"\d\d_.*", d) is None:
            continue
        ldir = os.path.join(lesson_dir, d)
        print("Scanning lesson dir {}".format(ldir))
        for lr, ld, lf in os.walk(ldir):
            for f in lf:
                if re.match(r".*\.ipynb", f) is None:
                    continue
                # We have a notebook!
                fpath = os.path.join(ldir, f)
                try:
                    print("Converting {} ...".format(fpath))
                    com = ["jupyter", "nbconvert", "--to", "html", fpath]
                    sp.check_call(com)
                    html = re.sub(r"\.ipynb", ".html", f)
                    infile = os.path.join(ldir, html)
                    outdir = os.path.join(nersc_dir, d)
                    outfile = os.path.join(outdir, html)
                    os.makedirs(outdir, exist_ok=True)
                    print("Move {} --> {}".format(infile, outfile))
                    shutil.copy2(infile, outfile)
                    os.chmod(
                        outfile,
                        stat.S_IROTH
                        | stat.S_IRGRP
                        | stat.S_IRUSR
                        | stat.S_IWGRP
                        | stat.S_IWUSR,
                    )
                    os.remove(infile)
                except:
                    print("Failed conversion")
                    raise
            break
    break



#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""Generate a focalplane pickle file compatible with the example scripts.
"""

import sys
import argparse
import traceback

import pickle

import numpy as np

from scipy.constants import degree

from toast.mpi import get_world

from toast.utils import Logger

from toast.tod import hex_pol_angles_qu, hex_layout, plot_focalplane


def main():
    log = Logger.get()

    parser = argparse.ArgumentParser(
        description="Simulate fake hexagonal focalplane.", fromfile_prefix_chars="@"
    )

    parser.add_argument(
        "--minpix",
        required=False,
        type=int,
        default=100,
        help="minimum number of pixels to use",
    )

    parser.add_argument(
        "--out",
        required=False,
        default="fp_fake",
        help="Root name of output pickle file",
    )

    parser.add_argument(
        "--fwhm", required=False, type=float, default=5.0, help="beam FWHM in arcmin"
    )

    parser.add_argument(
        "--fwhm_sigma",
        required=False,
        type=float,
        default=0,
        help="Relative beam FWHM distribution width",
    )

    parser.add_argument(
        "--fov",
        required=False,
        type=float,
        default=5.0,
        help="Field of View in degrees",
    )

    parser.add_argument(
        "--psd_fknee",
        required=False,
        type=float,
        default=0.05,
        help="Detector noise model f_knee in Hz",
    )

    parser.add_argument(
        "--psd_NET",
        required=False,
        type=float,
        default=60.0e-6,
        help="Detector noise model NET in K*sqrt(sec)",
    )

    parser.add_argument(
        "--psd_alpha",
        required=False,
        type=float,
        default=1.0,
        help="Detector noise model slope",
    )

    parser.add_argument(
        "--psd_fmin",
        required=False,
        type=float,
        default=1.0e-5,
        help="Detector noise model f_min in Hz",
    )

    parser.add_argument(
        "--bandcenter_ghz",
        required=False,
        type=float,
        help="Band center frequency [GHz]",
    )

    parser.add_argument(
        "--bandcenter_sigma",
        required=False,
        type=float,
        default=0,
        help="Relative band center distribution width",
    )

    parser.add_argument(
        "--bandwidth_ghz", required=False, type=float, help="Bandwidth [GHz]"
    )

    parser.add_argument(
        "--bandwidth_sigma",
        required=False,
        type=float,
        default=0,
        help="Relative bandwidth distribution width",
    )

    parser.add_argument(
        "--random_seed",
        required=False,
        type=np.int,
        default=123456,
        help="Random number generator seed for randomized " "detector parameters",
    )

    try:
        args = parser.parse_args()
    except SystemExit:
        return

    # Get the default communicator
    mpiworld, procs, rank = get_world()

    # Guard against being called with multiple processes
    if rank == 0:
        # Make one big hexagon layout at the center of the focalplane.
        # Compute the number of pixels that is at least the number requested.

        test = args.minpix - 1
        nrings = 0
        while (test - 6 * nrings) > 0:
            test -= 6 * nrings
            nrings += 1

        npix = 1
        for r in range(1, nrings + 1):
            npix += 6 * r

        log.info("using {} pixels ({} detectors)".format(npix, npix * 2))

        # Translate the field-of-view into distance between flag sides
        angwidth = args.fov * np.cos(30 * degree)

        Apol = hex_pol_angles_qu(npix, offset=0.0)
        Bpol = hex_pol_angles_qu(npix, offset=90.0)

        Adets = hex_layout(npix, angwidth, "fake_", "A", Apol)
        Bdets = hex_layout(npix, angwidth, "fake_", "B", Bpol)

        dets = Adets.copy()
        dets.update(Bdets)

        np.random.seed(args.random_seed)

        for indx, d in enumerate(sorted(dets.keys())):
            dets[d]["fknee"] = args.psd_fknee
            dets[d]["fmin"] = args.psd_fmin
            dets[d]["alpha"] = args.psd_alpha
            dets[d]["NET"] = args.psd_NET
            # This is in degrees, but the input is in arcmin.
            dets[d]["fwhm_deg"] = (args.fwhm / 60.0) * (
                1 + np.random.randn() * args.fwhm_sigma
            )
            # This is a fixed value, in arcmin.
            dets[d]["fwhm"] = args.fwhm
            if args.bandcenter_ghz:
                dets[d]["bandcenter_ghz"] = args.bandcenter_ghz * (
                    1 + np.random.randn() * args.bandcenter_sigma
                )
            if args.bandwidth_ghz:
                dets[d]["bandwidth_ghz"] = args.bandwidth_ghz * (
                    1 + np.random.randn() * args.bandwidth_sigma
                )
            dets[d]["index"] = indx

        outfile = "{}_{}".format(args.out, npix)
        qdets = {x: y["quat"] for x, y in dets.items()}
        beams = {x: (60.0 * y["fwhm_deg"]) for x, y in dets.items()}
        plot_focalplane(qdets, args.fov, args.fov, "{}.png".format(outfile), fwhm=beams)

        with open("{}.pkl".format(outfile), "wb") as p:
            pickle.dump(dets, p)

    return


if __name__ == "__main__":
    try:
        main()
    except:
        # We have an unhandled exception on at least one process.  Print a stack
        # trace for this process and then abort so that all processes terminate.
        mpiworld, procs, rank = get_world()
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(rank, x) for x in lines]
        print("".join(lines), flush=True)
        if mpiworld is not None:
            mpiworld.Abort(6)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os

import shutil

import numpy as np

import healpy as hp

from ..todmap import TODHpixSpiral, OpSimGradient, OpPointingHpix, OpMadam

from ._helpers import create_outdir, create_distdata, boresight_focalplane


class OpMadamTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # One observation per group
        self.data = create_distdata(self.comm, obs_per_group=1)

        self.ndet = self.data.comm.group_size
        self.rate = 50.0

        # Create detectors with defaults
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(self.ndet, samplerate=self.rate)

        # Samples per observation
        self.totsamp = 3 * 49152

        # Pixelization
        self.sim_nside = 64
        self.map_nside = 64

        # Populate the observations

        tod = TODHpixSpiral(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=self.data.comm.group_size,
            rate=self.rate,
            nside=self.sim_nside,
        )

        self.data.obs[0]["tod"] = tod

    def test_madam_gradient(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank

        # add simple sky gradient signal
        grad = OpSimGradient(nside=self.sim_nside, nest=True)
        grad.exec(self.data)

        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(self.data)

        # Write outputs to a test-specific directory
        mapdir = os.path.join(self.outdir, "grad")
        if rank == 0:
            if os.path.isdir(mapdir):
                shutil.rmtree(mapdir)
            os.makedirs(mapdir)

        handle = None
        if rank == 0:
            handle = open(os.path.join(mapdir, "out_test_madam_info"), "w")
        self.data.info(handle=handle)
        if rank == 0:
            handle.close()

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = min(8, self.map_nside)
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = mapdir
        pars["info"] = 1

        madam = OpMadam(params=pars, name="grad")
        if madam.available:
            # Run Madam twice on the same data and ensure the result
            # does not change
            madam.exec(self.data)

            m0 = None
            if rank == 0:
                m0 = hp.read_map(os.path.join(mapdir, "madam_bmap.fits"))

            madam.exec(self.data)

            m1 = None
            failed = False
            if rank == 0:
                m1 = hp.read_map(os.path.join(mapdir, "madam_bmap_001.fits"))
                if not np.allclose(m0, m1):
                    print("Madam did not produce the same map from the " "same data.")
                    failed = True
            if self.comm is not None:
                failed = self.comm.bcast(failed, root=0)
            self.assertFalse(failed)
        else:
            print("libmadam not available, skipping tests")

    def test_madam_output(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank

        # add simple sky gradient signal
        grad = OpSimGradient(nside=self.sim_nside, nest=True)
        grad.exec(self.data)

        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(self.data)

        # Write outputs to a test-specific directory
        mapdir = os.path.join(self.outdir, "out")
        if rank == 0:
            if os.path.isdir(mapdir):
                shutil.rmtree(mapdir)
            os.makedirs(mapdir)

        handle = None
        if rank == 0:
            handle = open(os.path.join(mapdir, "out_test_madam_info"), "w")
        self.data.info(handle=handle)
        if rank == 0:
            handle.close()

        pars = {}
        pars["kfirst"] = "T"
        pars["iter_max"] = 100
        pars["base_first"] = 5.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = min(8, self.map_nside)
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = mapdir
        pars["info"] = 0

        madam = OpMadam(params=pars, name="grad", name_out="destriped")

        if madam.available:
            tod = self.data.obs[0]["tod"]
            rms1 = None
            for det in tod.local_dets:
                ref_in = tod.cache.reference("grad_" + det)
                rms0 = np.std(ref_in)
                ref_in[ref_in.size // 2 :] += 1.0e6  # Add an offset
                rms1 = np.std(ref_in)
                del ref_in

            madam.exec(self.data)

            for det in tod.local_dets:
                ref_out = tod.cache.reference("destriped_" + det)
                rms2 = np.std(ref_out)
                del ref_out
                if rms1 < 0.9 * rms2:
                    raise Exception("Destriped TOD does not have lower RMS")
        else:
            print("libmadam not available, skipping tests")



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

# import functions in our public API

from .pysm import pysm

if pysm is not None:
    from .pysm import PySMSky

from .todmap_math import (
    OpAccumDiag,
    OpScanScale,
    OpScanMask,
    dipole,
)

from .pointing import OpPointingHpix, OpMuellerPointingHpix

from .sim_tod import (
    satellite_scanning,
    TODHpixSpiral,
    TODSatellite,
    slew_precession_axis,
    TODGround,
)

from .sim_det_map import OpSimGradient, OpSimScan

from .sim_det_dipole import OpSimDipole

from .sim_det_pysm import OpSimPySM

from .sim_det_atm import OpSimAtmosphere

from .sss import OpSimScanSynchronousSignal

from .groundfilter import OpGroundFilter

from .pointing_math import aberrate

from .conviqt import OpSimConviqt, OpSimWeightedConviqt

from .atm import available_utils as atm_available_utils
from .mapsampler import MapSampler

from .madam import OpMadam
from .mapmaker import OpMapMaker



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os
import time

import pickle

import numpy as np

from .mpi import MPITestCase

from ..timing import Timer, GlobalTimers, function_timer, gather_timers, dump

from ._helpers import create_outdir


@function_timer
def fibonacci(n):
    if n < 2:
        return n
    return fibonacci(n - 1) + fibonacci(n - 2)


class TimingTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

    def test_single(self):
        incr = 200
        dincr = float(incr) / 1000.0
        prec = 1
        tm = Timer()
        self.assertFalse(tm.is_running())
        tm.start()
        time.sleep(dincr)
        elapsed = tm.elapsed_seconds()
        tm.report_elapsed("Test timer elapsed")
        tm.stop()
        try:
            elapsed = tm.elapsed_seconds("This should raise since timer is stopped...")
        except:
            print("Successful exception:  elapsed_seconds() from a stopped timer")
        try:
            tm.report_elapsed("This should raise since timer is stopped...")
        except:
            print("Successful exception:  report_elapsed() for a stopped timer")
        np.testing.assert_almost_equal(tm.seconds(), elapsed, decimal=prec)
        np.testing.assert_almost_equal(tm.seconds(), dincr, decimal=prec)
        tm.report("Test timer stopped")
        tm.clear()
        tm.start()
        time.sleep(dincr)
        try:
            tm.report("This should raise since timer not stopped...")
        except:
            print("Successful exception:  report running timer")
        self.assertTrue(tm.is_running())
        tm.stop()
        # Test pickling
        pkl = pickle.dumps(tm, 2)
        newtm = pickle.loads(pkl)
        tm.report("original")
        newtm.report("pickle roundtrip")

    def test_global(self):
        incr = 200
        dincr = float(incr) / 1000.0
        prec = 2
        gt = GlobalTimers.get()
        tnames = ["timer1", "timer2", "timer3"]
        for nm in tnames:
            try:
                gt.stop(nm)
                tm.report("This should raise since timer not started...")
            except:
                print("Successful exception:  stop an already stopped timer")
        for nm in tnames:
            gt.start(nm)
        for nm in tnames:
            self.assertTrue(gt.is_running(nm))
            try:
                s = gt.seconds(nm)
            except:
                print("Successful exception:  seconds() on running timer")
        gt.stop_all()
        gt.clear_all()
        for nm in tnames:
            gt.start(nm)
        for nm in tnames:
            time.sleep(dincr)
            gt.stop(nm)
        offset = 1
        for nm in tnames:
            np.testing.assert_almost_equal(gt.seconds(nm), offset * dincr, decimal=prec)
            offset += 1
        gt.report()

    def test_comm(self):
        gt = GlobalTimers.get()
        for i in [10, 7, 11]:
            fibonacci(i - 2)
            fibonacci(i - 1)
            fibonacci(i)
        gt.stop_all()
        np = 1
        proc = 0
        if self.comm is not None:
            np = self.comm.size
            proc = self.comm.rank
        for p in range(np):
            if proc == p:
                print("--- Rank {}".format(p), flush=True)
                gt.report()
                print("---", flush=True)
            if self.comm is not None:
                self.comm.barrier()

        result = gather_timers(comm=self.comm)
        if proc == 0:
            for nm in sorted(result.keys()):
                print("{} timing:".format(nm), flush=True)
                props = result[nm]
                for k in sorted(props.keys()):
                    print("  {} = {}".format(k, props[k]), flush=True)
            out = os.path.join(self.outdir, "test_dump")
            dump(result, out)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import sys
import time

import warnings

from unittest.signals import registerResult
from unittest import TestCase
from unittest import TestResult


class MPITestCase(TestCase):
    """A simple wrapper around the standard TestCase which provides
    one extra method to set the communicator.
    """

    def __init__(self, *args, **kwargs):
        super(MPITestCase, self).__init__(*args, **kwargs)

    def setComm(self, comm):
        self.comm = comm


class MPITestResult(TestResult):
    """A test result class that can print formatted text results to a stream.

    The actions needed are coordinated across all processes.

    Used by MPITestRunner.
    """

    separator1 = "=" * 70
    separator2 = "-" * 70

    def __init__(self, comm, stream=None, descriptions=None, verbosity=None, **kwargs):
        super(MPITestResult, self).__init__(
            stream=stream, descriptions=descriptions, verbosity=verbosity, **kwargs
        )
        self.comm = comm
        self.stream = stream
        self.descriptions = descriptions
        self.buffer = False
        self.failfast = False

    def getDescription(self, test):
        doc_first_line = test.shortDescription()
        if self.descriptions and doc_first_line:
            return "\n".join((str(test), doc_first_line))
        else:
            return str(test)

    def startTest(self, test):
        if isinstance(test, MPITestCase):
            test.setComm(self.comm)
        self.stream.flush()
        if self.comm is not None:
            self.comm.barrier()
        if (self.comm is None) or (self.comm.rank == 0):
            self.stream.write("\n")
            self.stream.write(self.getDescription(test))
            self.stream.write(" ... ")
            self.stream.flush()
        if self.comm is not None:
            self.comm.barrier()
        super(MPITestResult, self).startTest(test)
        return

    def addSuccess(self, test):
        super(MPITestResult, self).addSuccess(test)
        if self.comm is None:
            self.stream.write("ok ")
        else:
            self.stream.write("[{}]ok ".format(self.comm.rank))
        self.stream.flush()
        return

    def addError(self, test, err):
        super(MPITestResult, self).addError(test, err)
        if self.comm is None:
            self.stream.write("error ")
        else:
            self.stream.write("[{}]error ".format(self.comm.rank))
        self.stream.flush()
        return

    def addFailure(self, test, err):
        super(MPITestResult, self).addFailure(test, err)
        if self.comm is None:
            self.stream.write("fail ")
        else:
            self.stream.write("[{}]fail ".format(self.comm.rank))
        self.stream.flush()
        return

    def addSkip(self, test, reason):
        super(MPITestResult, self).addSkip(test, reason)
        if self.comm is None:
            self.stream.write("skipped({}) ".format(reason))
        else:
            self.stream.write("[{}]skipped({}) ".format(self.comm.rank, reason))
        self.stream.flush()
        return

    def addExpectedFailure(self, test, err):
        super(MPITestResult, self).addExpectedFailure(test, err)
        if self.comm is None:
            self.stream.write("expected-fail ")
        else:
            self.stream.write("[{}]expected-fail ".format(self.comm.rank))
        self.stream.flush()
        return

    def addUnexpectedSuccess(self, test):
        super(MPITestResult, self).addUnexpectedSuccess(test)
        if self.comm is None:
            self.stream.writeln("unexpected-success ")
        else:
            self.stream.writeln("[{}]unexpected-success ".format(self.comm.rank))
        return

    def printErrorList(self, flavour, errors):
        for test, err in errors:
            if self.comm is None:
                self.stream.writeln("{}".format(self.separator1))
                self.stream.writeln("{}: {}".format(flavour, self.getDescription(test)))
                self.stream.writeln("{}".format(self.separator2))
                self.stream.writeln("{}".format(err))
            else:
                self.stream.writeln("[{}] {}".format(self.comm.rank, self.separator1))
                self.stream.writeln(
                    "[{}] {}: {}".format(
                        self.comm.rank, flavour, self.getDescription(test)
                    )
                )
                self.stream.writeln("[{}] {}".format(self.comm.rank, self.separator2))
                self.stream.writeln("[{}] {}".format(self.comm.rank, err))
        return

    def printErrors(self):
        if self.comm is None:
            self.stream.writeln()
            self.stream.flush()
            self.printErrorList("ERROR", self.errors)
            self.printErrorList("FAIL", self.failures)
            self.stream.flush()
        else:
            self.comm.barrier()
            if self.comm.rank == 0:
                self.stream.writeln()
                self.stream.flush()
            for p in range(self.comm.size):
                if p == self.comm.rank:
                    self.printErrorList("ERROR", self.errors)
                    self.printErrorList("FAIL", self.failures)
                    self.stream.flush()
                self.comm.barrier()
        return

    def allSuccessful(self):
        mysuccess = self.wasSuccessful()
        total = 0
        if not mysuccess:
            total = 1
        alltotal = None
        if self.comm is None:
            alltotal = total
        else:
            alltotal = self.comm.allreduce(total)
        if alltotal == 0:
            return True
        else:
            return False


class _WritelnDecorator(object):
    """Used to decorate file-like objects with a handy "writeln" method"""

    def __init__(self, stream):
        self.stream = stream

    def __getattr__(self, attr):
        if attr in ("stream", "__getstate__"):
            raise AttributeError(attr)
        return getattr(self.stream, attr)

    def writeln(self, arg=None):
        if arg:
            self.write(arg)
        self.write("\n")  # text-mode streams translate to \r\n if needed


class MPITestRunner(object):
    """A test runner class that displays results in textual form.

    It prints out the names of tests as they are run, errors as they
    occur, and a summary of the results at the end of the test run.

    This information is only printed by the root process.
    """

    resultclass = MPITestResult

    def __init__(
        self, comm, stream=None, descriptions=True, verbosity=2, warnings=None
    ):
        """Construct a MPITestRunner.

        Subclasses should accept **kwargs to ensure compatibility as the
        interface changes.
        """
        self.comm = comm
        if stream is None:
            stream = sys.stderr
        self.stream = _WritelnDecorator(stream)
        self.descriptions = descriptions
        self.verbosity = verbosity
        self.warnings = warnings

    def run(self, test):
        "Run the given test case or test suite."
        result = MPITestResult(
            self.comm, self.stream, self.descriptions, self.verbosity
        )
        registerResult(result)
        with warnings.catch_warnings():
            if self.warnings:
                # if self.warnings is set, use it to filter all the warnings
                warnings.simplefilter(self.warnings)
                # if the filter is "default" or "always", special-case the
                # warnings from the deprecated unittest methods to show them
                # no more than once per module, because they can be fairly
                # noisy.  The -Wd and -Wa flags can be used to bypass this
                # only when self.warnings is None.
                if self.warnings in ["default", "always"]:
                    warnings.filterwarnings(
                        "module",
                        category=DeprecationWarning,
                        message=r"Please use assert\w+ instead.",
                    )
            startTime = time.time()
            startTestRun = getattr(result, "startTestRun", None)
            if startTestRun is not None:
                startTestRun()
            try:
                test(result=result)
            finally:
                stopTestRun = getattr(result, "stopTestRun", None)
                if stopTestRun is not None:
                    stopTestRun()
            stopTime = time.time()
        self.stream.flush()
        if self.comm is None:
            self.stream.write("\n")
            self.stream.flush()
        else:
            self.comm.barrier()
            if self.comm.rank == 0:
                self.stream.write("\n")
                self.stream.flush()
        timeTaken = stopTime - startTime

        result.printErrors()

        if (self.comm is None) or (self.comm.rank == 0):
            if hasattr(result, "separator2"):
                self.stream.writeln(result.separator2)

        run = result.testsRun
        if (self.comm is None) or (self.comm.rank == 0):
            self.stream.writeln(
                "Ran %d test%s in %.3fs" % (run, run != 1 and "s" or "", timeTaken)
            )
            self.stream.writeln()

        expectedFails = unexpectedSuccesses = skipped = 0
        try:
            results = map(
                len,
                (result.expectedFailures, result.unexpectedSuccesses, result.skipped),
            )
        except AttributeError:
            pass
        else:
            expectedFails, unexpectedSuccesses, skipped = results

        psize = 1
        prank = 0
        if self.comm is not None:
            psize = self.comm.size
            prank = self.comm.rank

        for p in range(psize):
            if p == prank:
                infos = []
                if not result.wasSuccessful():
                    if self.comm is None:
                        self.stream.write("FAILED")
                    else:
                        self.stream.write("[{}] FAILED".format(self.comm.rank))
                    failed, errored = len(result.failures), len(result.errors)
                    if failed:
                        infos.append("failures=%d" % failed)
                    if errored:
                        infos.append("errors=%d" % errored)
                else:
                    if self.comm is None:
                        self.stream.write("OK")
                    else:
                        self.stream.write("[{}] OK".format(self.comm.rank))
                if skipped:
                    infos.append("skipped=%d" % skipped)
                if expectedFails:
                    infos.append("expected failures=%d" % expectedFails)
                if unexpectedSuccesses:
                    infos.append("unexpected successes=%d" % unexpectedSuccesses)
                if infos:
                    self.stream.writeln(" ({})".format(", ".join(infos)))
                else:
                    self.stream.write("\n")
                self.stream.flush()
            if self.comm is not None:
                self.comm.barrier()

        # if not result.allSuccessful():
        #     self.comm.Abort(1)

        return result



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os

import numpy as np

from ..tod import OpGainScrambler, AnalyticNoise, OpSimNoise
from ..todmap import TODHpixSpiral

from ._helpers import (
    create_outdir,
    create_distdata,
    boresight_focalplane,
    uniform_chunks,
)


class OpGainScramblerTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # One observation per group
        self.data = create_distdata(self.comm, obs_per_group=1)

        self.ndet = 4
        self.rate = 20.0

        # Create detectors with a range of knee frequencies.
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(
            self.ndet,
            samplerate=self.rate,
            net=10.0,
            fmin=1.0e-5,
            fknee=np.linspace(0.0, 0.1, num=self.ndet),
        )

        # Total samples per observation
        self.totsamp = 200000

        # Chunks
        chunks = uniform_chunks(self.totsamp, nchunk=self.data.comm.group_size)

        # Construct an empty TOD (no pointing needed)

        tod = TODHpixSpiral(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            firsttime=0.0,
            rate=self.rate,
            nside=512,
            sampsizes=chunks,
        )

        # construct an analytic noise model

        nse = AnalyticNoise(
            rate=drate,
            fmin=dfmin,
            detectors=dnames,
            fknee=dfknee,
            alpha=dalpha,
            NET=dnet,
        )

        self.data.obs[0]["tod"] = tod
        self.data.obs[0]["noise"] = nse

    def test_scrambler(self):
        # generate timestreams
        op = OpSimNoise()
        op.exec(self.data)

        # Record the old RMS
        old_rms = []
        for ob in self.data.obs:
            tod = ob["tod"]
            orms = {}
            for det in tod.local_dets:
                cachename = "noise_{}".format(det)
                y = tod.cache.reference(cachename)
                orms[det] = np.std(y)
                del y
            old_rms.append(orms)

        # Scramble the timestreams

        op = OpGainScrambler(center=2, sigma=1e-6, name="noise")
        op.exec(self.data)

        # Ensure RMS changes for the implicated detectors

        for ob, orms in zip(self.data.obs, old_rms):
            tod = ob["tod"]
            for det in tod.local_dets:
                cachename = "noise_{}".format(det)
                y = tod.cache.reference(cachename)
                rms = np.std(y)
                old = orms[det]
                if np.abs(rms / old) - 2 > 1e-3:
                    raise RuntimeError(
                        "det {} old rms = {}, new rms = {}".format(det, old, rms)
                    )
                del y

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from ..op import Operator

from ..timing import function_timer


class Interval(object):
    """Class storing a time and sample range.

    Args:
        start (float): The start time of the interval in seconds.
        stop (float): The stop time of the interval in seconds.
        first (int): The first sample index of the interval.
        last (int): The last sample index (inclusive) of the interval.

    """

    def __init__(self, start=None, stop=None, first=None, last=None):
        self._start = start
        self._stop = stop
        self._first = first
        self._last = last

    def __repr__(self):
        return "<Interval {} - {} ({} - {})>".format(
            self._start, self._stop, self._first, self._last
        )

    @property
    def start(self):
        """(float): the start time of the interval."""
        if self._start is None:
            raise RuntimeError("Start time is not yet assigned")
        return self._start

    @start.setter
    def start(self, val):
        if val < 0.0:
            raise ValueError("Negative start time is not valid")
        self._start = val

    @property
    def stop(self):
        """(float): the start time of the interval."""
        if self._stop is None:
            raise RuntimeError("Stop time is not yet assigned")
        return self._stop

    @stop.setter
    def stop(self, val):
        if val < 0.0:
            raise ValueError("Negative stop time is not valid")
        self._stop = val

    @property
    def first(self):
        """(int): the first sample of the interval."""
        if self._first is None:
            raise RuntimeError("First sample is not yet assigned")
        return self._first

    @first.setter
    def first(self, val):
        if val < 0:
            raise ValueError("Negative first sample is not valid")
        self._first = val

    @property
    def last(self):
        """(int): the first sample of the interval."""
        if self._last is None:
            raise RuntimeError("Last sample is not yet assigned")
        return self._last

    @last.setter
    def last(self, val):
        if val < 0:
            raise ValueError("Negative last sample is not valid")
        self._last = val

    @property
    def range(self):
        """(float): the number seconds in the interval."""
        b = self.start
        e = self.stop
        return e - b

    @property
    def samples(self):
        """(int): the number samples in the interval."""
        b = self.first
        e = self.last
        return e - b + 1


class OpFlagGaps(Operator):
    """Operator which applies common flags to gaps between valid intervals.

    Args:
        common_flag_name (str): the name of the cache object
            to use for the common flags.  If None, use the TOD.
        common_flag_value (int): the integer bit mask (0-255)
            that should be bitwise ORed with the existing flags.
        intervals (str):  Name of the valid intervals in observation.

    """

    def __init__(
        self, common_flag_name=None, common_flag_value=1, intervals="intervals"
    ):
        self._common_flag_name = common_flag_name
        self._common_flag_value = common_flag_value
        self._intervals = intervals
        super().__init__()

    @function_timer
    def exec(self, data):
        """Flag samples between valid intervals.

        This iterates over all observations and flags samples
        which lie outside the list of intervals.

        Args:
            data (toast.Data): The distributed data.

        """
        for obs in data.obs:
            tod = obs["tod"]
            if (self._intervals is not None) and (self._intervals in obs):
                intervals = obs[self._intervals]
            else:
                intervals = None
            local_intervals = tod.local_intervals(intervals)

            # first, flag all samples
            offset, nsamp = tod.local_samples
            gapflags = np.zeros(nsamp, dtype=np.uint8)
            gapflags.fill(self._common_flag_value)

            # now un-flag samples in valid intervals
            for ival in local_intervals:
                local_start = ival.first
                local_stop = ival.last
                gapflags[local_start : local_stop + 1] = 0

            commonflags = tod.local_common_flags(self._common_flag_name)
            commonflags[:] |= gapflags

        return


def intervals_to_chunklist(intervals, nsamp, startsamp=0):
    """Create a list of contiguous sample chunks from intervals.

    Given a list of (possibly discontinuous) intervals, construct a
    list of contiguous chunks of samples.  The chunks are defined between
    the starting points of each interval.  An additional chunk at the
    beginning and end are added if necessary so that the sum of chunks
    equals the total number of samples.

    Args:
        intervals (list): sorted list of Interval objects.
        nsamp (int): the number of samples to consider.
        startsamp (int): the first sample to consider.

    Returns:
        (list): list of sample chunks.

    """
    chunks = list()
    previous = None
    for it in intervals:
        if it.last < startsamp:
            continue
        if it.first >= (startsamp + nsamp):
            continue
        if previous is None:
            # We are at the first interval which overlaps our
            # sample range.
            if it.first <= startsamp:
                previous = startsamp
            else:
                chunks.append(it.first - startsamp)
                previous = it.first
        else:
            chunks.append(it.first - previous)
            previous = it.first
    # Handle final chunk
    chunks.append(intervals[-1].last - previous + 1)
    sm = np.sum(chunks)
    if sm < nsamp:
        chunks.append(nsamp - sm)
    return chunks



#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""
This script creates a CES schedule file that can be used as input
to toast_ground_sim.py
"""

from toast.timing import GlobalTimers
from toast.scheduler import run_scheduler


def main():
    gt = GlobalTimers.get()
    gt.start("toast_ground_schedule")

    run_scheduler()

    gt.stop_all()
    gt.report()
    return


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        # We have an unhandled exception on at least one process.  Print a stack
        # trace for this process and then abort so that all processes terminate.
        mpiworld, procs, rank = get_world()
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(rank, x) for x in lines]
        print("".join(lines), flush=True)
        if mpiworld is not None and procs > 1:
            mpiworld.Abort(6)
        else:
            raise e



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os
import shutil

import numpy as np
import numpy.testing as nt

from .. import qarray as qa

from ..tod import regular_intervals, TODGround

from ._helpers import create_outdir, create_distdata, boresight_focalplane

# This file will only be imported if TIDAS is already available
import tidas as tds
from tidas.mpi import MPIVolume
from ..tod import tidas as tt
from ..tod import tidas_utils as ttutils


class TidasTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        self.outvol = os.path.join(self.outdir, "test_tidas")
        self.export = os.path.join(self.outdir, "export_tidas")
        self.groundexport = os.path.join(self.outdir, "export_tidas_ground")

        # Create observations, divided evenly between groups
        self.nobs = 4
        opg = self.nobs
        if (self.comm is not None) and (self.comm.size >= 2):
            opg = self.nobs // 2
        self.data = create_distdata(self.comm, obs_per_group=opg)

        # Create a set of boresight detectors
        self.ndet = 8
        self.rate = 20.0

        (
            self.dnames,
            self.dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(self.ndet, samplerate=self.rate)

        # Create a set of intervals that sub-divide each observation.

        self.obslen = 90.0
        self.obsgap = 10.0

        self.obstotalsamp = int((self.obslen + self.obsgap) * self.rate)
        self.obssamp = int(self.obslen * self.rate)
        self.obsgapsamp = self.obstotalsamp - self.obssamp

        self.obstotal = (self.obstotalsamp - 1) / self.rate
        self.obslen = (self.obssamp - 1) / self.rate
        self.obsgap = (self.obsgapsamp - 1) / self.rate

        # Properties of the intervals within an observation
        self.nsub = 12
        self.subtotsamp = self.obssamp // self.nsub
        self.subgapsamp = 0
        self.subsamp = self.subtotsamp - self.subgapsamp

        # Ground scan properties
        self.site_lon = "-67:47:10"
        self.site_lat = "-22:57:30"
        self.site_alt = 5200.0
        self.coord = "C"
        self.azmin = 45
        self.azmax = 55
        self.el = 60
        self.scanrate = 1.0
        self.scan_accel = 0.1
        self.CES_start = None

    def tearDown(self):
        pass

    def meta_setup(self):
        ret = tds.Dictionary()
        ret.put_string("string", "blahblahblah")
        ret.put_float64("double", -1.234567890123e9)
        ret.put_float32("float", -123456.0)
        ret.put_int8("int8", -1)
        ret.put_uint8("uint8", 1)
        ret.put_int16("int16", -10000)
        ret.put_uint16("uint16", 10000)
        ret.put_int32("int32", -1000000000)
        ret.put_uint32("uint32", 1000000000)
        ret.put_int64("int64", -100000000000)
        ret.put_uint64("uint64", 100000000000)
        return ttutils.to_dict(ret)

    def meta_verify(self, dct):
        nt.assert_equal(dct.get_string("string"), "blahblahblah")
        nt.assert_equal(dct.get_int8("int8"), -1)
        nt.assert_equal(dct.get_uint8("uint8"), 1)
        nt.assert_equal(dct.get_int16("int16"), -10000)
        nt.assert_equal(dct.get_uint16("uint16"), 10000)
        nt.assert_equal(dct.get_int32("int32"), -1000000000)
        nt.assert_equal(dct.get_uint32("uint32"), 1000000000)
        nt.assert_equal(dct.get_int64("int64"), -100000000000)
        nt.assert_equal(dct.get_uint64("uint64"), 100000000000)
        nt.assert_almost_equal(dct.get_float32("float"), -123456.0)
        nt.assert_almost_equal(dct.get_float64("double"), -1.234567890123e9)
        return

    def create_intervals(self, start, first):
        subtime = self.subsamp / self.rate
        subgap = self.subgapsamp / self.rate

        ret = regular_intervals(self.nsub, start, first, self.rate, subtime, subgap)

        tret = [tds.Intrvl(x.start, x.stop, x.first, x.last) for x in ret]

        return tret

    def intervals_init(self, start, first):
        return self.create_intervals(start, first)

    def intervals_verify(self, ilist, start, first):
        reg = self.create_intervals(start, first)
        inlen = int(len(ilist))
        checklen = int(len(reg))
        self.assertTrue(inlen == checklen)

        for i in range(len(ilist)):
            nt.assert_almost_equal(ilist[i].start, reg[i].start)
            nt.assert_equal(ilist[i].first, reg[i].first)
        return

    def create_bore(self, total, local):
        theta_incr = (0.5 * np.pi) / (total - 1)
        phi_incr = (2.0 * np.pi) / (total - 1)

        theta_start = local[0] * theta_incr
        phi_start = local[0] * phi_incr

        theta_stop = theta_start + (local[1] - 1) * theta_incr
        phi_stop = phi_start + (local[1] - 1) * phi_incr

        theta = np.linspace(
            theta_start, theta_stop, num=local[1], endpoint=True, dtype=np.float64
        )
        phi = np.linspace(
            phi_start, phi_stop, num=local[1], endpoint=True, dtype=np.float64
        )
        pa = np.zeros(local[1], dtype=np.float64)

        return qa.from_angles(theta, phi, pa)

    def obs_create(self, vol, name, obsid, start, first):
        # create the observation within the TIDAS volume
        tob = ttutils.find_obs(vol, "", name)
        obsprops = tds.Dictionary()
        obsprops.put_int64("obs_id", obsid)
        obsprops.put_int64("obs_telescope_id", 1234)
        obsprops.put_int64("obs_site_id", 5678)

        # Create the observation group
        obsgroup = tob.group_add("observation", tds.Group(tds.Schema(), obsprops, 0))
        del obsprops
        del obsgroup

        # create intervals
        ilist = self.intervals_init(start, 0)
        dist = [(y.first - x.first) for x, y in zip(ilist[:-1], ilist[1:])]
        dist.append(self.obstotalsamp - ilist[-1].first)

        intr = tob.intervals_add("chunks", tds.Intervals(tds.Dictionary(), len(ilist)))
        # print("rank {}: write intervals".format(self.comm.rank), flush=True)
        intr.write(ilist)
        del intr
        del tob

        # Create a TOD for this observation.
        tt.TODTidas.create(
            vol,
            "/{}".format(name),
            self.dquat,
            self.obstotalsamp,
            self.meta_setup(),
            True,
        )

        return

    def obs_init(self, obscomm, vol, name, start, first):
        detranks = 1
        if obscomm is not None:
            detranks = obscomm.size
        tod = tt.TODTidas(
            obscomm, detranks, vol, "/{}".format(name), distintervals="chunks"
        )

        # print("rank {}: done create TOD {}".format(self.comm.rank, name), flush=True)

        # Now write the data.  For this test, we simply write the detector
        # index (as a float) to the detector timestream.  We also flag every
        # other sample.  For the boresight pointing, we create a fake spiral
        # pattern.

        detranks, sampranks = tod.grid_size
        rankdet, ranksamp = tod.grid_ranks

        off = tod.local_samples[0]
        n = tod.local_samples[1]

        # We use this for both the common and all the detector
        # flags just to check write/read roundtrip.
        flags = np.zeros(n, dtype=np.uint8)
        flags[::2] = 1

        if rankdet == 0:
            # print("rank {}: working on timestamps".format(self.comm.rank), flush=True)
            # Processes along the first row of the process grid write
            # their slice of the timestamps.
            incr = 1.0 / self.rate
            stamps = np.arange(n, dtype=np.float64)
            stamps *= incr
            stamps += start + (off * incr)

            for p in range(sampranks):
                if p == ranksamp:
                    # print("rank {}: write timestamp slice".format(self.comm.rank), flush=True)
                    tod.write_times(stamps=stamps)
                # print("rank {}: grid row barrier".format(self.comm.rank), flush=True)
                if tod.grid_comm_row is not None:
                    tod.grid_comm_row.barrier()

            # print("rank {}: done timestamps".format(self.comm.rank), flush=True)

            # Same with the boresight
            boresight = self.create_bore(tod.total_samples, tod.local_samples)

            for p in range(sampranks):
                if p == ranksamp:
                    # print("rank {}: write boresight".format(self.comm.rank), flush=True)
                    tod.write_boresight(data=boresight)
                # print("rank {}: bore grid row barrier".format(self.comm.rank), flush=True)
                if tod.grid_comm_row is not None:
                    tod.grid_comm_row.barrier()

            # Now the common flags
            for p in range(sampranks):
                if p == ranksamp:
                    # print("rank {}: write common".format(self.comm.rank), flush=True)
                    tod.write_common_flags(flags=flags)
                # print("rank {}: com grid row barrier".format(self.comm.rank), flush=True)
                if tod.grid_comm_row is not None:
                    tod.grid_comm_row.barrier()

        # print("rank {}: hit tod comm barrier".format(self.comm.rank), flush=True)
        if tod.mpicomm is not None:
            tod.mpicomm.barrier()
        # print("rank {}: done tod comm barrier".format(self.comm.rank), flush=True)

        # Detector data- serialize for now
        groupsize = 1
        grouprank = 0
        if tod.mpicomm is not None:
            groupsize = tod.mpicomm.size
            grouprank = tod.mpicomm.rank

        for p in range(groupsize):
            if p == grouprank:
                fakedata = np.empty(n, dtype=np.float64)
                for d in tod.local_dets:
                    # get unique detector index and convert to float
                    indx = float(tod.detindx[d])
                    # write this to all local elements
                    fakedata[:] = np.arange(tod.local_samples[1])
                    fakedata += tod.local_samples[0]
                    fakedata *= indx
                    # print("rank {}: write det {}".format(self.comm.rank, d), flush=True)
                    tod.write(detector=d, data=fakedata)
                    # write detector flags
                    tod.write_flags(detector=d, flags=flags)
            # print("rank {}: write det barrier".format(self.comm.rank), flush=True)
            if tod.mpicomm is not None:
                tod.mpicomm.barrier()

        del tod

        return

    def obs_verify(self, vol, parent, name, tod, start, first, ignore_last):
        nlocal = tod.local_samples[1]
        odd = False
        if tod.local_samples[0] % 2 != 0:
            odd = True

        block = ttutils.find_obs(vol, parent, name)

        # Read the intervals and compare

        intr = block.intervals_get("chunks")
        ilist = intr.read()
        if ignore_last:
            # These intervals included a final one representing the last
            # gap in sampling.  Trim it before comparing.
            ilist = ilist[0:-1]
        self.intervals_verify(ilist, start, 0)

        # Verify group properties

        self.meta_verify(ttutils.from_dict(tod.meta()))

        # Read and compare timestamps

        compstamps = np.arange(nlocal, dtype=np.float64)
        compstamps /= self.rate
        compstamps += start + (tod.local_samples[0] / self.rate)

        stamps = tod.read_times()

        nt.assert_almost_equal(stamps, compstamps)

        # Read and compare boresight

        compbore = self.create_bore(tod.total_samples, tod.local_samples)
        boresight = tod.read_boresight()

        nt.assert_almost_equal(boresight, compbore)

        # flags.  We use this for both the common and all the detector
        # flags just to check write/read roundtrip.

        compflags = np.zeros(nlocal, dtype=np.uint8)
        if odd:
            compflags[1::2] = 1
        else:
            compflags[::2] = 1

        flags = tod.read_common_flags()

        nt.assert_equal(flags, compflags)

        # detector data

        compdata = np.empty(nlocal, dtype=np.float64)
        for d in tod.local_dets:
            # get unique detector index and convert to float
            indx = float(tod.detindx[d])
            # comparison values
            compdata[:] = np.arange(tod.local_samples[1])
            compdata += tod.local_samples[0]
            compdata *= indx
            # read and check
            data = tod.read(detector=d)
            nt.assert_almost_equal(data, compdata)
            # check detector flags
            flags = tod.read_flags(detector=d)
            nt.assert_equal(flags, compflags)

        return

    def volume_init(self, path):
        rank = 0
        grouprank = 0
        if self.comm is not None:
            rank = self.comm.rank
            grouprank = self.data.comm.group_rank
        if rank == 0:
            if os.path.isdir(path):
                shutil.rmtree(path)
        if self.comm is not None:
            self.comm.barrier()

        # print("rank {}: creating volume".format(self.comm.rank), flush=True)
        # Create the volume on the world communicator
        vol = None
        if self.comm is None:
            vol = tds.Volume(
                path, tds.BackendType.hdf5, tds.CompressionType.none, dict()
            )
        else:
            vol = MPIVolume(
                self.data.comm.comm_world,
                path,
                tds.BackendType.hdf5,
                tds.CompressionType.none,
                dict(),
            )
        del vol
        # print("rank {}: creating volume done".format(self.comm.rank), flush=True)
        # Usually for real data we will have a hierarchy of blocks
        # (year, month, season, etc).  For this test we just add generic
        # observations to the root block.

        # For simplicity, just have every group take turns writing their data.
        # this is fine for a small unit test.

        for pgrp in range(self.data.comm.ngroups):
            if pgrp == self.data.comm.group:
                vol = None
                if self.comm is None:
                    vol = tds.Volume(path, tds.AccessMode.write)
                else:
                    vol = MPIVolume(
                        self.data.comm.comm_group, path, tds.AccessMode.write
                    )

                # One process from each group creates the observations
                for ob in range(len(self.data.obs)):
                    obsname = "obs_{}_{:02d}".format(self.data.comm.group, ob)
                    start = ob * self.obstotal
                    first = ob * self.obstotalsamp
                    if grouprank == 0:
                        self.obs_create(vol, obsname, ob, start, first)

                if self.comm is not None:
                    vol.meta_sync()

                for ob in range(len(self.data.obs)):
                    obsname = "obs_{}_{:02d}".format(self.data.comm.group, ob)
                    start = ob * self.obstotal
                    first = ob * self.obstotalsamp
                    # print("rank {}: init obs {}".format(self.comm.rank, obsname), flush=True)
                    self.obs_init(self.data.comm.comm_group, vol, obsname, start, first)

                if self.comm is not None:
                    vol.meta_sync()
                del vol

            if self.comm is not None:
                self.comm.barrier()

        return

    def init_ground(self):
        # Create the simulated TODs
        for ob in range(len(self.data.obs)):
            obsname = "obs_{}_{:02d}".format(self.data.comm.group, ob)
            start = ob * self.obstotal
            first = ob * self.obstotalsamp
            tod = TODGround(
                self.data.comm.comm_group,
                self.dquat,
                self.obstotalsamp,
                detranks=self.data.comm.group_size,
                firsttime=start,
                rate=self.rate,
                site_lon=self.site_lon,
                site_lat=self.site_lat,
                site_alt=self.site_alt,
                azmin=self.azmin,
                azmax=self.azmax,
                el=self.el,
                coord=self.coord,
                scanrate=self.scanrate,
                scan_accel=self.scan_accel,
                CES_start=self.CES_start,
            )
            self.data.obs[ob]["tod"] = tod
        return

    def volume_verify(self, path, detgroup="detectors", ignore_last=False):
        vol = None
        if self.comm is None:
            vol = tds.Volume(path, tds.AccessMode.read)
        else:
            vol = MPIVolume(self.data.comm.comm_world, path, tds.AccessMode.read)
        root = vol.root()

        for ob in range(len(self.data.obs)):
            obsname = "obs_{}_{:02d}".format(self.data.comm.group, ob)
            # print("rank {}: verify obs {}".format(self.comm.rank, obsname), flush=True)
            obs = root.block_get(obsname)
            start = ob * self.obstotal
            first = ob * self.obstotalsamp
            ilist = self.intervals_init(start, 0)
            dist = [(y.first - x.first) for x, y in zip(ilist[:-1], ilist[1:])]
            dist.append(self.obstotalsamp - ilist[-1].first)
            tod = tt.TODTidas(
                self.data.comm.comm_group,
                self.data.comm.group_size,
                vol,
                "/{}".format(obsname),
                distintervals=dist,
                group_dets=detgroup,
            )
            self.obs_verify(vol, "", obsname, tod, start, first, ignore_last)
        del vol
        return

    def test_io(self):
        self.volume_init(self.outvol)
        self.volume_verify(self.outvol)
        return

    def test_export(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        self.volume_init(self.outvol)

        distdata = tt.load_tidas(
            self.data.comm,
            self.data.comm.group_size,
            self.outvol,
            "w",
            "detectors",
            tt.TODTidas,
            distintervals="chunks",
        )

        if rank == 0:
            if os.path.isdir(self.export):
                shutil.rmtree(self.export)
        if self.comm is not None:
            self.comm.barrier()

        # NOTE:  the use_todchunks option will take the distribution chunks
        # of the TOD and create a set of intervals in the volume.  These
        # chunks were themselves created from a set of intervals above with
        # gaps.  So there will be final chunk that is the last "gap" period
        # to add up to the total samples.
        dumper = tt.OpTidasExport(
            self.export, tt.TODTidas, backend="hdf5", use_todchunks=True
        )
        dumper.exec(distdata)

        self.volume_verify(self.export, ignore_last=True)

        # print("finished verify 1", flush=True)
        distdata2 = tt.load_tidas(
            self.data.comm,
            self.data.comm.group_size,
            self.export,
            "r",
            "detectors",
            tt.TODTidas,
            group_dets="detectors",
            distintervals="chunks",
        )

        # print("finished load 2", flush=True)

        dumper2 = tt.OpTidasExport(
            self.export,
            tt.TODTidas,
            use_todchunks=True,
            create_opts={"group_dets": "bolos"},
            ctor_opts={"group_dets": "bolos"},
        )
        dumper2.exec(distdata2)

        # print("finished export 2", flush=True)

        self.volume_verify(self.export, detgroup="bolos", ignore_last=True)

        return

    def test_ground(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # Create simulated tods in memory
        self.init_ground()

        # Export this
        if rank == 0:
            if os.path.isdir(self.groundexport):
                shutil.rmtree(self.groundexport)
        if self.comm is not None:
            self.comm.barrier()

        dumper = tt.OpTidasExport(
            self.groundexport, tt.TODTidas, backend="hdf5", use_todchunks=True
        )
        dumper.exec(self.data)

        # Load it back in.
        distdata = tt.load_tidas(
            self.data.comm,
            self.data.comm.group_size,
            self.groundexport,
            "w",
            "detectors",
            tt.TODTidas,
            distintervals="chunks",
            group_dets="detectors",
        )

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from ..mpi import MPI, use_mpi

import os
import sys
import unittest

from .mpi import MPITestRunner

from ..vis import set_backend

from .._libtoast import libtoast_tests

from . import env as testenv
from . import cache as testcache
from . import timing as testtiming
from . import rng as testrng
from . import fft as testfft
from . import healpix as testhealpix
from . import dist as testdist
from . import qarray as testqarray
from . import tod as testtod

from . import psd_math as testpsdmath
from . import intervals as testintervals

from . import cov as testcov

from . import ops_pmat as testopspmat

from . import ops_dipole as testopsdipole
from . import ops_simnoise as testopssimnoise
from . import ops_sim_sss as testopssimsss

from . import ops_polyfilter as testopspolyfilter
from . import ops_groundfilter as testopsgroundfilter

from . import ops_gainscrambler as testopsgainscrambler
from . import ops_applygain as testopsapplygain

from . import ops_memorycounter as testopsmemorycounter

from . import ops_madam as testopsmadam
from . import ops_mapmaker as testopsmapmaker

from . import map_satellite as testmapsatellite

from . import map_ground as testmapground

from . import binned as testbinned

from . import sim_focalplane as testsimfocalplane
from . import tod_satellite as testtodsat

from ..todmap import pysm

if pysm is not None:
    from . import ops_sim_pysm as testopspysm

from . import ops_sim_atm as testopsatm

from ..tod import tidas_available

# if tidas_available:
#     from . import tidas as testtidas
testtidas = None
tidas_available = False

# from ..tod import spt3g_available
# if spt3g_available:
#     from . import spt3g as testspt3g
testspt3g = None
spt3g_available = False


def test(name=None, verbosity=2):
    # We run tests with COMM_WORLD if available
    comm = None
    rank = 0
    if use_mpi:
        comm = MPI.COMM_WORLD
        rank = comm.rank

    set_backend()

    outdir = "toast_test_output"

    if rank == 0:
        outdir = os.path.abspath(outdir)
        if not os.path.isdir(outdir):
            os.makedirs(outdir)

    if comm is not None:
        outdir = comm.bcast(outdir, root=0)

    if (name is None) or (name == "libtoast"):
        # Run tests from the serial compiled library.
        libtoast_tests(list(sys.argv))

    # Run python tests.

    loader = unittest.TestLoader()
    mpirunner = MPITestRunner(comm, verbosity=verbosity, warnings="ignore")
    suite = unittest.TestSuite()

    if name is None:
        suite.addTest(loader.loadTestsFromModule(testenv))
        suite.addTest(loader.loadTestsFromModule(testcache))
        if not (("CONDA_BUILD" in os.environ) or ("CIBUILDWHEEL" in os.environ)):
            # When doing a conda build on CI services in containers
            # the timing information is not accurate and these tests
            # fail.
            suite.addTest(loader.loadTestsFromModule(testtiming))
        suite.addTest(loader.loadTestsFromModule(testrng))
        suite.addTest(loader.loadTestsFromModule(testfft))
        suite.addTest(loader.loadTestsFromModule(testhealpix))
        suite.addTest(loader.loadTestsFromModule(testdist))
        suite.addTest(loader.loadTestsFromModule(testqarray))
        suite.addTest(loader.loadTestsFromModule(testtod))
        suite.addTest(loader.loadTestsFromModule(testtodsat))
        suite.addTest(loader.loadTestsFromModule(testintervals))
        suite.addTest(loader.loadTestsFromModule(testopssimnoise))
        suite.addTest(loader.loadTestsFromModule(testopssimsss))
        suite.addTest(loader.loadTestsFromModule(testopsapplygain))
        suite.addTest(loader.loadTestsFromModule(testopspmat))
        suite.addTest(loader.loadTestsFromModule(testcov))
        suite.addTest(loader.loadTestsFromModule(testopsdipole))
        suite.addTest(loader.loadTestsFromModule(testopsgroundfilter))
        suite.addTest(loader.loadTestsFromModule(testsimfocalplane))
        suite.addTest(loader.loadTestsFromModule(testopspolyfilter))
        suite.addTest(loader.loadTestsFromModule(testopsmemorycounter))
        suite.addTest(loader.loadTestsFromModule(testopsgainscrambler))
        suite.addTest(loader.loadTestsFromModule(testpsdmath))
        suite.addTest(loader.loadTestsFromModule(testopsmadam))
        suite.addTest(loader.loadTestsFromModule(testopsmapmaker))
        suite.addTest(loader.loadTestsFromModule(testmapsatellite))
        suite.addTest(loader.loadTestsFromModule(testmapground))
        suite.addTest(loader.loadTestsFromModule(testbinned))
        suite.addTest(loader.loadTestsFromModule(testopsatm))
        # These tests segfault locally.  Re-enable once we are doing bandpass
        # integration on on the fly.
        # if pysm is not None:
        #     suite.addTest(loader.loadTestsFromModule(testopspysm))

        if tidas_available:
            suite.addTest(loader.loadTestsFromModule(testtidas))
        if spt3g_available:
            suite.addTest(loader.loadTestsFromModule(testspt3g))
    elif name != "libtoast":
        if (name == "tidas") and (not tidas_available):
            print("Cannot run TIDAS tests- package not available")
            return
        elif (name == "spt3g") and (not spt3g_available):
            print("Cannot run SPT3G tests- package not available")
            return
        else:
            modname = "toast.tests.{}".format(name)
            if modname not in sys.modules:
                result = '"{}" is not a valid test.  Try'.format(name)
                for name in sys.modules:
                    if name.startswith("toast.tests."):
                        result += '\n  - "{}"'.format(name.replace("toast.tests.", ""))
                result += "\n"
                raise RuntimeError(result)
            suite.addTest(loader.loadTestsFromModule(sys.modules[modname]))

    ret = 0
    _ret = mpirunner.run(suite)
    if not _ret.wasSuccessful():
        ret += 1

    if ret > 0:
        sys.exit(ret)

    return ret



##
# Copyright (c) 2017-2020, all rights reserved.  Use of this source code
# is governed by a BSD license that can be found in the top-level
# LICENSE file.
##

import os
import sys
import time

import unittest

import numpy as np
import numpy.testing as nt

from .shmem import MPIShared
from .locking import MPILock

MPI = None
use_mpi = True

if "PSHMEM_MPI_DISABLE" in os.environ:
    use_mpi = False

if use_mpi and (MPI is None):
    try:
        import mpi4py.MPI as MPI
    except ImportError:
        raise ImportError("Cannot import mpi4py, will only test serial functionality.")


class ShmemTest(unittest.TestCase):
    def setUp(self):
        self.comm = None
        if MPI is not None:
            self.comm = MPI.COMM_WORLD
        self.rank = 0
        self.procs = 1
        if self.comm is not None:
            self.rank = self.comm.rank
            self.procs = self.comm.size

    def tearDown(self):
        pass

    def test_allocate(self):
        # Dimensions of our shared memory array
        datadims = (2, 5, 10)

        # Dimensions of the incremental slab that we will
        # copy during each set() call.
        updatedims = (1, 1, 5)

        # How many updates are there to cover the whole
        # data array?
        nupdate = 1
        for d in range(len(datadims)):
            nupdate *= datadims[d] // updatedims[d]

        for datatype in [np.int32, np.int64, np.float32, np.float64]:

            # For testing the "set()" method, every process is going to
            # create a full-sized data buffer and fill it with its process rank.
            local = np.ones(datadims, dtype=datatype)
            local *= self.rank

            # A context manager is the pythonic way to make sure that the
            # object has no dangling reference counts after leaving the context,
            # and will ensure that the shared memory is freed properly.

            with MPIShared(local.shape, local.dtype, self.comm) as shm:
                for p in range(self.procs):
                    # Every process takes turns writing to the buffer.
                    setdata = None
                    setoffset = (0, 0, 0)

                    # Write to the whole data volume, but in small blocks
                    for upd in range(nupdate):
                        if p == self.rank:
                            # My turn!  Write my process rank to the buffer slab.
                            setdata = local[
                                setoffset[0] : setoffset[0] + updatedims[0],
                                setoffset[1] : setoffset[1] + updatedims[1],
                                setoffset[2] : setoffset[2] + updatedims[2],
                            ]
                        try:
                            # All processes call set(), but only data on rank p matters.
                            shm.set(setdata, setoffset, fromrank=p)
                        except:
                            print(
                                "proc {} threw exception during set()".format(rank),
                                flush=True,
                            )
                            if self.comm is not None:
                                self.comm.Abort()
                            else:
                                sys.exit(1)

                        # Increment the write offset within the array

                        x = setoffset[0]
                        y = setoffset[1]
                        z = setoffset[2]

                        z += updatedims[2]
                        if z >= datadims[2]:
                            z = 0
                            y += updatedims[1]
                        if y >= datadims[1]:
                            y = 0
                            x += updatedims[0]

                        setoffset = (x, y, z)

                    # Every process is now going to read a copy from the shared memory
                    # and make sure that they see the data written by the current process.
                    check = np.zeros_like(local)
                    check[:, :, :] = shm[:, :, :]

                    truth = np.ones_like(local)
                    truth *= p

                    # This should be bitwise identical, even for floats
                    nt.assert_equal(check[:, :, :], truth[:, :, :])

                # Ensure that we can reference the memory buffer from numpy without
                # a memory copy.  The intention is that a slice of the shared memory
                # buffer should appear as a C-contiguous ndarray whenever we slice
                # along the last dimension.

                for p in range(self.procs):
                    if p == self.rank:
                        slc = shm[1, 2]
                        print(
                            "proc {} slice has dims {}, dtype {}, C = {}".format(
                                p, slc.shape, slc.dtype.str, slc.flags["C_CONTIGUOUS"]
                            ),
                            flush=True,
                        )
                    if self.comm is not None:
                        self.comm.barrier()

    def test_shape(self):
        good_dims = [
            (2, 5, 10),
            np.array([10, 2], dtype=np.int32),
            np.array([5, 2], dtype=np.int64),
            np.array([10, 2], dtype=np.int),
        ]
        bad_dims = [
            (2, 5.5, 10),
            np.array([10, 2], dtype=np.float32),
            np.array([5, 2], dtype=np.float64),
            np.array([10, 2.5], dtype=np.float32),
        ]

        dt = np.float64

        for dims in good_dims:
            try:
                shm = MPIShared(dims, dt, self.comm)
                if self.rank == 0:
                    print("successful creation with shape {}".format(dims), flush=True)
                del shm
            except ValueError:
                if self.rank == 0:
                    print(
                        "unsuccessful creation with shape {}".format(dims), flush=True
                    )
        for dims in bad_dims:
            try:
                shm = MPIShared(dims, dt, self.comm)
                if self.rank == 0:
                    print("unsuccessful rejection of shape {}".format(dims), flush=True)
                del shm
            except ValueError:
                if self.rank == 0:
                    print("successful rejection of shape {}".format(dims), flush=True)


class LockTest(unittest.TestCase):
    def setUp(self):
        self.comm = None
        if MPI is not None:
            self.comm = MPI.COMM_WORLD
        self.rank = 0
        self.procs = 1
        if self.comm is not None:
            self.rank = self.comm.rank
            self.procs = self.comm.size
        self.sleepsec = 0.2

    def tearDown(self):
        pass

    def test_lock(self):
        with MPILock(self.comm, root=0, debug=True) as lock:
            for lk in range(5):
                msg = "test_lock:  process {} got lock {}".format(self.rank, lk)
                lock.lock()
                print(msg, flush=True)
                # time.sleep(self.sleepsec)
                lock.unlock()
        if self.comm is not None:
            self.comm.barrier()


def run():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(LockTest))
    suite.addTest(unittest.makeSuite(ShmemTest))
    runner = unittest.TextTestRunner()
    runner.run(suite)
    return



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from ..tod import OpGainScrambler


def add_gainscrambler_args(parser):
    """Add the noise simulation arguments"""
    parser.add_argument(
        "--gainscrambler",
        required=False,
        default=False,
        action="store_true",
        help="Add simulated noise",
        dest="apply_gainscrambler",
    )
    parser.add_argument(
        "--no-gainscrambler",
        required=False,
        action="store_false",
        help="Do not add simulated noise",
        dest="apply_gainscrambler",
    )

    parser.add_argument(
        "--gain-sigma",
        required=False,
        default=0.01,
        type=np.float,
        help="Simulated gain fluctuation amplitude",
    )
    return


@function_timer
def scramble_gains(args, comm, data, mc, cache_name=None, verbose=False):
    if not args.apply_gainscrambler or args.gain_sigma == 0:
        return
    log = Logger.get()
    timer = Timer()
    timer.start()
    if comm.world_rank == 0 and verbose:
        log.info("Scrambling gains")
    scrambler = OpGainScrambler(sigma=args.gain_sigma, name=cache_name, realization=mc)
    scrambler.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    timer.stop()
    if comm.world_rank == 0 and verbose:
        timer.report("Scramble gains")
    return



# This setup.py file simply builds TOAST using the underlying cmake build
# system.  This is only preferred in certain cases where the automation is
# easier from a setup.py (e.g. readthedocs, pip, etc).

import os
import sys
import re
import subprocess
import glob
from pathlib import Path

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext


def find_compilers():
    cc = None
    cxx = None
    # If we have mpi4py, then get the MPI compilers that were used to build that.
    # Then get the serial compilers used by the MPI wrappers.  Otherwise, just the
    # normal distutils compilers.
    try:
        from mpi4py import MPI
        import mpi4py

        mpiconf = mpi4py.get_config()
        mpicc = mpiconf["mpicc"]
        mpicxx = mpiconf["mpicxx"]
        mpicc_com = None
        mpicxx_com = None
        try:
            mpicc_com = subprocess.check_output(
                "{} -show".format(mpicc), shell=True, universal_newlines=True
            )
        except CalledProcessError:
            # Cannot run the MPI C compiler, give up
            raise ImportError
        try:
            mpicxx_com = subprocess.check_output(
                "{} -show".format(mpicxx), shell=True, universal_newlines=True
            )
        except CalledProcessError:
            # Cannot run the MPI C++ compiler, give up
            raise ImportError
        # Extract the serial compilers
        cc = mpicc_com.split()[0]
        cxx = mpicxx_com.split()[0]

    except ImportError:
        pass

    return (cc, cxx)


def get_version():
    # Run the underlying cmake command that generates the version file, and then
    # parse that output.  This way setup.py is using the exact same version as the
    # (not yet built) compiled code.
    topdir = Path(__file__).resolve().parent
    ver = None
    try:
        version_dir = os.path.join(topdir, "src", "libtoast")
        # version_dir = os.path.join("src", "libtoast")
        subprocess.check_call("cmake -P version.cmake", shell=True, cwd=version_dir)
        version_cpp = os.path.join(version_dir, "version.cpp")
        git_ver = None
        rel_ver = None
        with open(version_cpp, "r") as f:
            for line in f:
                mat = re.match(r'.*GIT_VERSION = "(.*)".*', line)
                if mat is not None:
                    git_ver = mat.group(1)
                mat = re.match(r'.*RELEASE_VERSION = "(.*)".*', line)
                if mat is not None:
                    rel_ver = mat.group(1)
        if (git_ver is not None) and (git_ver != ""):
            ver = git_ver
        else:
            ver = rel_ver
    except CalledProcessError:
        raise RuntimeError("Cannot generate version!")
    return ver


class CMakeExtension(Extension):
    """
    This overrides the built-in extension class and essentially does nothing,
    since all extensions are compiled in one go by the custom build_ext class.
    """

    def __init__(self, name, sources=[]):
        super().__init__(name=name, sources=sources)


class CMakeBuild(build_ext):
    """
    Builds the full package using CMake.
    """

    def run(self):
        """
        Perform build_cmake before doing the 'normal' stuff
        """
        for extension in self.extensions:
            if extension.name == "toast._libtoast":
                # We always build the serial extension, so we trigger the build
                # on that.
                self.build_cmake()
        super().run()

    def build_cmake(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        (cc, cxx) = find_compilers()

        # Make a copy of the environment so that we can modify it
        env = os.environ.copy()

        # Search the environment for any variables starting with "TOAST_BUILD_".
        # We extract these and convert them into cmake options.
        cmake_opts = dict()
        cpat = re.compile(r"TOAST_BUILD_(.*)")
        for k, v in env.items():
            mat = cpat.match(k)
            if mat is not None:
                cmake_opts[mat.group(1)] = v

        cmake_args = ["-DPYTHON_EXECUTABLE=" + sys.executable]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]

        # Set compilers

        if cc is not None:
            # Use serial compilers that were used when building MPI
            cmake_args += ["-DCMAKE_C_COMPILER={}".format(cc)]
        elif "CMAKE_C_COMPILER" in cmake_opts:
            # Get these from the environment
            cmake_args += [
                "-DCMAKE_C_COMPILER={}".format(cmake_opts["CMAKE_C_COMPILER"])
            ]
            _ = cmake_opts.pop("CMAKE_C_COMPILER")
        else:
            # We just let cmake guess the compilers and hope for the best...
            pass

        if cxx is not None:
            # Use serial compilers that were used when building MPI
            cmake_args += ["-DCMAKE_CXX_COMPILER={}".format(cxx)]
        elif "CMAKE_CXX_COMPILER" in cmake_opts:
            # Get these from the environment
            cmake_args += [
                "-DCMAKE_CXX_COMPILER={}".format(cmake_opts["CMAKE_CXX_COMPILER"])
            ]
            _ = cmake_opts.pop("CMAKE_CXX_COMPILER")
        else:
            # We just let cmake guess the compilers and hope for the best...
            pass

        # Append any other TOAST_BUILD_ options to the cmake args
        for k, v in cmake_opts.items():
            cmake_args += ["-D{}={}".format(k, v)]

        # Assuming Makefiles
        build_args += ["--", "-j2"]

        self.build_args = build_args

        env["CXXFLAGS"] = "{} -DVERSION_INFO=\\'{}\\'".format(
            env.get("CXXFLAGS", ""), get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        # CMakeLists.txt is in the same directory as this setup.py file
        cmake_list_dir = os.path.abspath(os.path.dirname(__file__))
        print("-" * 10, "Running CMake prepare", "-" * 40)
        subprocess.check_call(
            ["cmake", cmake_list_dir] + cmake_args, cwd=self.build_temp, env=env
        )

        print("-" * 10, "Building extensions", "-" * 40)
        cmake_cmd = ["cmake", "--build", "."] + self.build_args
        subprocess.check_call(cmake_cmd, cwd=self.build_temp)

        # Move from build temp to final position
        for ext in self.extensions:
            self.move_output(ext)

    def move_output(self, ext):
        extpath = self.get_ext_filename(ext.name)
        build_temp = Path(self.build_temp).resolve()
        dest_path = Path(self.get_ext_fullpath(ext.name)).resolve()
        source_path = os.path.join(build_temp, "src", extpath)
        dest_directory = dest_path.parents[0]
        dest_directory.mkdir(parents=True, exist_ok=True)
        self.copy_file(source_path, dest_path)


ext_modules = [CMakeExtension("toast._libtoast")]

scripts = glob.glob("pipelines/*.py")


def readme():
    with open("README.md") as f:
        return f.read()


conf = dict()
conf["name"] = "toast-cmb"
conf["description"] = "Time Ordered Astrophysics Scalable Tools"
conf["long_description"] = readme()
conf["long_description_content_type"] = "text/markdown"
conf["author"] = "Theodore Kisner, Reijo Keskitalo"
conf["author_email"] = "tskisner.public@gmail.com"
conf["license"] = "BSD"
conf["url"] = "https://github.com/hpc4cmb/toast"
conf["version"] = get_version()
conf["python_requires"] = ">=3.6.0"
conf["install_requires"] = [
    "cmake",
    "numpy",
    "scipy",
    "healpy",
    "matplotlib",
    "ephem",
    "h5py",
]
conf["extras_require"] = {"mpi": ["mpi4py>=3.0"]}
conf["packages"] = find_packages("src")
conf["package_dir"] = {"": "src"}
conf["ext_modules"] = ext_modules
conf["scripts"] = scripts
conf["cmdclass"] = {"build_ext": CMakeBuild}
conf["zip_safe"] = False
conf["classifiers"] = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Operating System :: POSIX",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering :: Astronomy",
]

setup(**conf)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os
import shutil

import numpy as np
import numpy.testing as nt

import healpy as hp

from .. import qarray as qa

from ..tod import AnalyticNoise, OpSimNoise
from ..todmap import (
    TODSatellite,
    OpPointingHpix,
    OpSimGradient,
    slew_precession_axis,
    satellite_scanning,
    OpSimScan,
    OpMadam,
)

from ..map import DistPixels

from ._helpers import (
    create_outdir,
    create_distdata,
    boresight_focalplane,
    uniform_chunks,
)


class MapSatelliteTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group, and each observation will have
        # one detector with multiple chunks distributed among the processes.

        self.data = create_distdata(self.comm, obs_per_group=1)
        self.data_fast_hwp = create_distdata(self.comm, obs_per_group=1)
        self.data_const_hwp = create_distdata(self.comm, obs_per_group=1)

        self.ndet = 1
        self.rate = 40.0

        # Create detectors with white noise
        self.NET = 7.0

        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(
            self.ndet, samplerate=self.rate, net=self.NET, fknee=0.0
        )

        # Samples per observation
        self.totsamp = 4000000

        # Pixelization
        self.sim_nside = 64
        self.map_nside = 64
        self.nside_submap = 16

        # Scan strategy
        self.spinperiod = 10.0
        self.spinangle = 30.0
        self.precperiod = 50.0
        self.precangle = 65.0

        # in order to make sure that the noise realization is reproducible
        # at all concurrencies, we set the chunksize to something independent
        # of the number of ranks.

        # Chunks
        chunks = uniform_chunks(self.totsamp, nchunk=100)

        # Populate the single observation per group

        common_args = [dquat, self.totsamp]
        common_kwargs = {
            "detranks": 1,
            "firsttime": 0.0,
            "rate": self.rate,
            "spinperiod": self.spinperiod,
            "spinangle": self.spinangle,
            "precperiod": self.precperiod,
            "precangle": self.precangle,
            "sampsizes": chunks,
        }

        # No HWP
        tod_no_hwp = TODSatellite(
            self.data.comm.comm_group, *common_args, **common_kwargs
        )

        # CHWP
        hwprate = self.rate * 60.0
        tod_fast_hwp = TODSatellite(
            self.data_fast_hwp.comm.comm_group,
            *common_args,
            **common_kwargs,
            hwprpm=hwprate
        )

        # Stepped HWP
        hwpstep = 2.0 * np.pi
        hwpsteptime = (self.totsamp / self.rate) / 60.0
        tod_const_hwp = TODSatellite(
            self.data_const_hwp.comm.comm_group,
            *common_args,
            **common_kwargs,
            hwpstep=hwpstep,
            hwpsteptime=hwpsteptime
        )

        for data, tod in [
            (self.data, tod_no_hwp),
            (self.data_fast_hwp, tod_fast_hwp),
            (self.data_const_hwp, tod_const_hwp),
        ]:
            precquat = np.empty(4 * tod.local_samples[1], dtype=np.float64).reshape(
                (-1, 4)
            )

            slew_precession_axis(
                precquat,
                firstsamp=tod.local_samples[0],
                samplerate=self.rate,
                degday=1.0,
            )

            tod.set_prec_axis(qprec=precquat)

            # add analytic noise model with white noise

            nse = AnalyticNoise(
                rate=drate,
                fmin=dfmin,
                detectors=dnames,
                fknee=dfknee,
                alpha=dalpha,
                NET=dnet,
            )

            data.obs[0]["tod"] = tod
            data.obs[0]["noise"] = nse

    def test_boresight_null(self):
        # verify that if all angles are zero, we get fixed pointing
        # along the x-axis.

        nsim = 1000

        zaxis = np.array([0, 0, 1], dtype=np.float64)

        borequat = np.empty(4 * nsim, dtype=np.float64).reshape((-1, 4))

        satellite_scanning(
            borequat,
            qprec=None,
            samplerate=100.0,
            spinperiod=1.0,
            spinangle=0.0,
            precperiod=20.0,
            precangle=0.0,
        )

        data = qa.rotate(borequat, np.tile(zaxis, nsim).reshape(-1, 3))

        np.testing.assert_almost_equal(
            data, np.tile(np.array([1.0, 0.0, 0.0]), nsim).reshape(-1, 3)
        )
        return

    def test_grad(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # add simple sky gradient signal
        grad = OpSimGradient(nside=self.sim_nside, nest=True)
        grad.exec(self.data)

        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(self.data)

        handle = None
        if rank == 0:
            handle = open(
                os.path.join(self.outdir, "out_test_satellite_grad_info"), "w"
            )
        self.data.info(handle=handle)
        if rank == 0:
            handle.close()

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_grad")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = 8
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(params=pars, name="grad", purge=True)
        if madam.available:
            madam.exec(self.data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to input signal

                tothits = np.sum(hits)
                nt.assert_equal(self.data.comm.ngroups * self.totsamp, tothits)

                sig = grad.sigmap()
                mask = bins > -1.0e20
                nt.assert_almost_equal(bins[mask], sig[mask], decimal=4)
        else:
            print("libmadam not available, skipping tests")
        return

    def test_noise(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # generate noise timestreams from the noise model
        nsig = OpSimNoise()
        nsig.exec(self.data)

        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(self.data)

        handle = None
        if rank == 0:
            handle = open(
                os.path.join(self.outdir, "out_test_satellite_noise_info"), "w"
            )
        self.data.info(handle=handle)
        if rank == 0:
            handle.close()

        # For noise weighting in madam, we know we are using an analytic noise
        # and so we can use noise weights based on the NET.  This is instrument
        # specific.

        tod = self.data.obs[0]["tod"]
        nse = self.data.obs[0]["noise"]
        detweights = {}
        for d in tod.local_dets:
            detweights[d] = 1.0 / (self.rate * nse.NET(d) ** 2)

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_noise")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = 8
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(params=pars, detweights=detweights, name="noise")
        if madam.available:
            madam.exec(self.data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # check that pixel rms makes sense given the
                # number of hits and the timestream rms

                tothits = np.sum(hits)
                nt.assert_equal(self.data.comm.ngroups * self.totsamp, tothits)
                # print("tothits = ", tothits)

                mask = bins > -1.0e20
                # print("num good pix = ", len(mask))
                rthits = np.sqrt(hits[mask].astype(np.float64))
                # print("rthits = ", rthits)
                # print("bmap = ", bins[mask])
                weighted = bins[mask] * rthits
                pixrms = np.std(weighted)
                todrms = self.NET * np.sqrt(self.rate)
                relerr = np.absolute(pixrms - todrms) / todrms
                # print("pixrms = ", pixrms)
                # print("todrms = ", todrms)
                # print("relerr = ", relerr)
                self.assertTrue(relerr < 0.05)
        else:
            print("libmadam not available, skipping tests")
        return

    def test_scanmap(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(self.data)

        # construct a sky gradient operator, just to get the signal
        # map- we are not going to use the operator on the data.
        grad = OpSimGradient(nside=self.sim_nside, nest=True)
        sig = grad.sigmap()

        # construct a distributed map which has the gradient
        distsig = DistPixels(
            self.data, comm=self.data.comm.comm_group, nnz=1, dtype=np.float64
        )

        distsig.broadcast_healpix_map(sig)

        # create TOD from map
        scansim = OpSimScan(input_map=distsig)
        scansim.exec(self.data)

        handle = None
        if rank == 0:
            handle = open(
                os.path.join(self.outdir, "out_test_satellite_scanmap_info"), "w"
            )
        self.data.info(handle=handle)
        if rank == 0:
            handle.close()

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_scansim")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = 8
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(params=pars, name="scan")
        if madam.available:
            madam.exec(self.data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to input signal

                tothits = np.sum(hits)
                nt.assert_equal(self.data.comm.ngroups * self.totsamp, tothits)
                mask = bins > -1.0e20
                nt.assert_almost_equal(bins[mask], sig[mask], decimal=4)
        else:
            print("libmadam not available, skipping tests")
        return

    def test_hwpfast(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # make a pointing matrix with a HWP that rotates 2*PI every sample
        data = self.data_fast_hwp
        pointing = OpPointingHpix(nside=self.map_nside, nest=True)
        pointing.exec(data)

        # construct a sky gradient operator, just to get the signal
        # map- we are not going to use the operator on the data.
        grad = OpSimGradient(nside=self.sim_nside, nest=True)
        sig = grad.sigmap()

        # construct a distributed map which has the gradient
        distsig = DistPixels(data, comm=data.comm.comm_group, nnz=1, dtype=np.float64)

        distsig.broadcast_healpix_map(sig)

        # create TOD from map
        scansim = OpSimScan(input_map=distsig)
        scansim.exec(data)

        handle = None
        if rank == 0:
            handle = open(
                os.path.join(self.outdir, "out_test_satellite_hwpfast_info"), "w"
            )
        data.info(handle=handle)
        if rank == 0:
            handle.close()

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_hwpfast")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = 8
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(params=pars, name="scan")
        if madam.available:
            madam.exec(data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to input signal

                tothits = np.sum(hits)
                nt.assert_equal(data.comm.ngroups * self.totsamp, tothits)
                mask = bins > -1.0e20
                nt.assert_almost_equal(bins[mask], sig[mask], decimal=4)
        else:
            print("libmadam not available, skipping tests")
        return

    def test_hwpconst(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # make a pointing matrix with a HWP that is constant
        data = self.data_const_hwp
        pointing = OpPointingHpix(
            nside=self.map_nside, nest=True, nside_submap=self.nside_submap
        )
        pointing.exec(data)

        # construct a sky gradient operator, just to get the signal
        # map- we are not going to use the operator on the data.
        grad = OpSimGradient(nside=self.sim_nside, nest=True)
        sig = grad.sigmap()

        # construct a distributed map which has the gradient
        distsig = DistPixels(data, comm=data.comm.comm_group, nnz=1, dtype=np.float64)

        distsig.broadcast_healpix_map(sig)

        # create TOD from map
        scansim = OpSimScan(input_map=distsig)
        scansim.exec(data)

        handle = None
        if rank == 0:
            handle = open(
                os.path.join(self.outdir, "out_test_satellite_hwpconst_info"), "w"
            )
        data.info(handle=handle)
        if rank == 0:
            handle.close()

        # make a binned map with madam
        madam_out = os.path.join(self.outdir, "madam_hwpconst")
        if rank == 0:
            if os.path.isdir(madam_out):
                shutil.rmtree(madam_out)
            os.mkdir(madam_out)

        pars = {}
        pars["kfirst"] = "F"
        pars["base_first"] = 1.0
        pars["fsample"] = self.rate
        pars["nside_map"] = self.map_nside
        pars["nside_cross"] = self.map_nside
        pars["nside_submap"] = 8
        pars["write_map"] = "F"
        pars["write_binmap"] = "T"
        pars["write_matrix"] = "F"
        pars["write_wcov"] = "F"
        pars["write_hits"] = "T"
        pars["kfilter"] = "F"
        pars["path_output"] = madam_out
        pars["info"] = 0

        madam = OpMadam(params=pars, name="scan")
        if madam.available:
            madam.exec(data)

            if rank == 0:
                import matplotlib.pyplot as plt

                hitsfile = os.path.join(madam_out, "madam_hmap.fits")
                hits = hp.read_map(hitsfile, nest=True)

                outfile = "{}.png".format(hitsfile)
                hp.mollview(hits, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                binfile = os.path.join(madam_out, "madam_bmap.fits")
                bins = hp.read_map(binfile, nest=True)

                outfile = "{}.png".format(binfile)
                hp.mollview(bins, xsize=1600, nest=True)
                plt.savefig(outfile)
                plt.close()

                # compare binned map to input signal

                tothits = np.sum(hits)
                nt.assert_equal(data.comm.ngroups * self.totsamp, tothits)
                mask = bins > -1.0e20
                nt.assert_almost_equal(bins[mask], sig[mask], decimal=4)
        else:
            print("libmadam not available, skipping tests")
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

import scipy.interpolate as si
from scipy.signal import fftconvolve

from ..op import Operator

from ..timing import function_timer

from ..utils import Logger, AlignedF64

from .. import rng as rng

from .._libtoast import tod_sim_noise_timestream


class OpCacheInit(Operator):
    """This operator initializes cache objects with the specified value"""

    def __init__(self, init_val=0, name=None):
        """
        Args:
            init_val (float) :  initial value to use
            name (str) :  cache prefix to use, can be None
        """
        self.init_val = init_val
        self.name = name

    def exec(self, data):
        """Create cache objects as <self.name>_<detector> and initialize
        them to self.init_val
        """
        for obs in data.obs:
            tod = obs["tod"]
            offset, nsamp = tod.local_samples
            for det in tod.local_dets:
                try:
                    # Initialize an existing cache object
                    tod.local_signal(det, self.name)[:] = self.init_val
                except:
                    # Object does not exist yet, create
                    cachename = "{}_{}".format(self.name, det)
                    tod.cache.put(
                        cachename,
                        np.full(nsamp, self.init_val, dtype=np.float64),
                        replace=True,
                    )
        return


class OpFlagsApply(Operator):
    """This operator sets the flagged signal values to zero"""

    def __init__(
        self, name=None, common_flags=None, flags=None, common_flag_mask=1, flag_mask=1
    ):
        """
        Args:
            name (str) :  cache prefix to apply the flags
            common_flags (str) :  cache_name to use for common flags
            flags (str) :  cache prefix to use for detector flags
            common_flag_mask (uint8) :  bit pattern to check the common
                flags against
            flag_mask (uint8) :  bit pattern to check the detector flags
                 against
        """
        self.name = name
        self.common_flags = common_flags
        self.flags = flags
        self.common_flag_mask = common_flag_mask
        self.flag_mask = flag_mask

    def exec(self, data):
        for obs in data.obs:
            tod = obs["tod"]
            common_flags = tod.local_common_flags(self.common_flags)
            common_flags = (common_flags & self.common_flag_mask) != 0
            for det in tod.local_dets:
                flags = tod.local_flags(det, self.flags)
                flags = (flags & self.flag_mask) != 0
                flags[common_flags] = True
                signal = tod.local_signal(det, self.name)
                signal[flags] = 0
        return


@function_timer
def calibrate(toitimes, toi, gaintimes, gains, order=0, inplace=False):
    """Interpolate the gains to TOI samples and apply them.

    Args:
        toitimes (float): Increasing TOI sample times in same units as
            gaintimes
        toi (float): TOI samples to calibrate
        gaintimes (float): Increasing timestamps of the gain values in
            same units as toitimes
        gains (float): Multiplicative gains
        order (int): Gain interpolation order. 0 means steps at the gain
            times, all other are polynomial interpolations.
        inplace (bool): Overwrite input TOI.

    Returns:
        calibrated timestream.

    """
    if len(gaintimes) == 1:
        g = gains
    else:
        if order == 0:
            ind = np.searchsorted(gaintimes, toitimes, side="right") - 1
            g = gains[ind]
        else:
            if len(gaintimes) <= order:
                order = len(gaintimes) - 1
            p = np.polyfit(gaintimes, gains, order)
            g = np.polyval(p, toitimes)

    if inplace:
        toi_out = toi
    else:
        toi_out = np.zeros_like(toi)

    toi_out[:] = toi * g

    return toi_out


@function_timer
def sim_noise_timestream(
    realization,
    telescope,
    component,
    obsindx,
    detindx,
    rate,
    firstsamp,
    samples,
    oversample,
    freq,
    psd,
    py=False,
):
    """Generate a noise timestream, given a starting RNG state.

    Use the RNG parameters to generate unit-variance Gaussian samples
    and then modify the Fourier domain amplitudes to match the desired
    PSD.

    The RNG (Threefry2x64 from Random123) takes a "key" and a "counter"
    which each consist of two unsigned 64bit integers.  These four
    numbers together uniquely identify a single sample.  We construct
    those four numbers in the following way:

    key1 = realization * 2^32 + telescope * 2^16 + component
    key2 = obsindx * 2^32 + detindx
    counter1 = currently unused (0)
    counter2 = sample in stream

    counter2 is incremented internally by the RNG function as it calls
    the underlying Random123 library for each sample.

    Args:
        realization (int): the Monte Carlo realization.
        telescope (int): a unique index assigned to a telescope.
        component (int): a number representing the type of timestream
            we are generating (detector noise, common mode noise,
            atmosphere, etc).
        obsindx (int): the global index of this observation.
        detindx (int): the global index of this detector.
        rate (float): the sample rate.
        firstsamp (int): the start sample in the stream.
        samples (int): the number of samples to generate.
        oversample (int): the factor by which to expand the FFT length
            beyond the number of samples.
        freq (array): the frequency points of the PSD.
        psd (array): the PSD values.
        py (bool): if True, use a pure-python implementation.  This is useful
            for testing.  If True, also return the interpolated PSD.

    Returns:
        (array):  the noise timestream.  If py=True, returns a tuple of timestream,
            interpolated frequencies, and interpolated PSD.

    """
    tdata = None
    if py:
        fftlen = 2
        while fftlen <= (oversample * samples):
            fftlen *= 2
        npsd = fftlen // 2 + 1
        norm = rate * float(npsd - 1)

        interp_freq = np.fft.rfftfreq(fftlen, 1 / rate)
        if interp_freq.size != npsd:
            raise RuntimeError(
                "interpolated PSD frequencies do not have expected length"
            )

        # Ensure that the input frequency range includes all the frequencies
        # we need.  Otherwise the extrapolation is not well defined.

        if np.amin(freq) < 0.0:
            raise RuntimeError("input PSD frequencies should be >= zero")

        if np.amin(psd) < 0.0:
            raise RuntimeError("input PSD values should be >= zero")

        increment = rate / fftlen

        if freq[0] > increment:
            raise RuntimeError(
                "input PSD does not go to low enough frequency to "
                "allow for interpolation"
            )

        nyquist = rate / 2
        if np.abs((freq[-1] - nyquist) / nyquist) > 0.01:
            raise RuntimeError(
                "last frequency element does not match Nyquist "
                "frequency for given sample rate: {} != {}".format(freq[-1], nyquist)
            )

        # Perform a logarithmic interpolation.  In order to avoid zero values, we
        # shift the PSD by a fixed amount in frequency and amplitude.

        psdshift = 0.01 * np.amin(psd[(psd > 0.0)])
        freqshift = increment

        loginterp_freq = np.log10(interp_freq + freqshift)
        logfreq = np.log10(freq + freqshift)
        logpsd = np.log10(psd + psdshift)

        interp = si.interp1d(logfreq, logpsd, kind="linear", fill_value="extrapolate")

        loginterp_psd = interp(loginterp_freq)
        interp_psd = np.power(10.0, loginterp_psd) - psdshift

        # Zero out DC value

        interp_psd[0] = 0.0

        scale = np.sqrt(interp_psd * norm)

        # gaussian Re/Im randoms, packed into a complex valued array

        key1 = realization * 4294967296 + telescope * 65536 + component
        key2 = obsindx * 4294967296 + detindx
        counter1 = 0
        counter2 = firstsamp * oversample

        rngdata = rng.random(
            fftlen, sampler="gaussian", key=(key1, key2), counter=(counter1, counter2)
        ).array()

        fdata = np.zeros(npsd, dtype=np.complex)

        # Set the DC and Nyquist frequency imaginary part to zero
        fdata[0] = rngdata[0] + 0.0j
        fdata[-1] = rngdata[npsd - 1] + 0.0j

        # Repack the other values.
        fdata[1:-1] = rngdata[1 : npsd - 1] + 1j * rngdata[-1 : npsd - 1 : -1]

        # scale by PSD
        fdata *= scale

        # inverse FFT
        tdata = np.fft.irfft(fdata)

        # subtract the DC level- for just the samples that we are returning
        offset = (fftlen - samples) // 2

        DC = np.mean(tdata[offset : offset + samples])
        tdata[offset : offset + samples] -= DC
        return (tdata[offset : offset + samples], interp_freq, interp_psd)
    else:
        tdata = AlignedF64(samples)
        tod_sim_noise_timestream(
            realization,
            telescope,
            component,
            obsindx,
            detindx,
            rate,
            firstsamp,
            oversample,
            freq.astype(np.float64),
            psd.astype(np.float64),
            tdata,
        )
        return tdata.array()


class OpCacheCopy(Operator):
    """Operator which copies sets of timestreams between cache locations.

    This simply copies data from one set of per-detector cache objects to
    another set.  At some point we will likely move away from persistent
    caching of intermediate timestreams and this operator will become
    irrelevant.

    Args:
        in (str): use cache objects with name <in>_<detector>.
        out (str): copy data to the cache with name <out>_<detector>.
            If the named cache objects do not exist, then they are created.
        force (bool): force creating the target cache object.

    """

    def __init__(self, input, output, force=False):
        # Call the parent class constructor.
        super().__init__()
        self._in = input
        self._out = output
        self._force = force

    @function_timer
    def exec(self, data):
        """Copy timestreams.

        This iterates over all observations and detectors and copies cache
        objects whose names match the specified pattern.

        Args:
            data (toast.Data): The distributed data.

        """
        for obs in data.obs:
            tod = obs["tod"]
            for det in tod.local_dets:
                inref = tod.local_signal(det, self._in)
                outname = "{}_{}".format(self._out, det)
                outref = tod.cache.put(outname, inref, replace=self._force)
                del outref
                del inref
        return


class OpCacheClear(Operator):
    """Operator which destroys cache objects matching the given pattern.

    Args:
        name (str): use cache objects with name <name>_<detector>.

    """

    def __init__(self, name):
        # Call the parent class constructor.
        super().__init__()
        self._name = name

    @function_timer
    def exec(self, data):
        """Clear timestreams.

        This iterates over all observations and detectors and clears cache
        objects whose names match the specified pattern.

        Args:
            data (toast.Data): The distributed data.

        """
        for obs in data.obs:
            tod = obs["tod"]
            for det in tod.local_dets:
                # if the cache object exists, destroy it
                name = "{}_{}".format(self._name, det)
                if tod.cache.exists(name):
                    tod.cache.destroy(name)
        return


@function_timer
def flagged_running_average(
    signal, flag, wkernel, return_flags=False, downsample=False
):
    """Compute a running average considering only the unflagged samples.

    Args:
        signal (float)
        flag (bool)
        wkernel (int):  Running average width
        return_flags (bool):  If true, also return flags which are
            a subset of the input flags.
        downsample (bool):  If True, return a downsampled version of the
            filtered timestream.

    Returns:
        (array or tuple):  The filtered signal and optionally the flags.

    """
    if len(signal) != len(flag):
        raise Exception("Signal and flag lengths do not match.")

    bad = flag != 0
    masked_signal = signal.copy()
    masked_signal[bad] = 0

    good = np.ones(len(signal), dtype=np.float64)
    good[bad] = 0

    kernel = np.ones(wkernel, dtype=np.float64)

    filtered_signal = fftconvolve(masked_signal, kernel, mode="same")
    filtered_hits = fftconvolve(good, kernel, mode="same")

    hit = filtered_hits > 0.1
    nothit = np.logical_not(hit)

    filtered_signal[hit] /= filtered_hits[hit]
    filtered_signal[nothit] = 0

    if return_flags or downsample:
        filtered_flags = np.zeros_like(flag)
        filtered_flags[nothit] = True

    if downsample:
        good = filtered_flags == 0
        if return_flags:
            filtered_flags[good][::wkernel]
        filtered_signal[good][::wkernel]

    if return_flags:
        return filtered_signal, filtered_flags
    else:
        return filtered_signal



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os

import numpy as np
import numpy.testing as nt

from .mpi import MPITestCase

from ..tod import sim_focalplane as sfp


def generate_hex(npix, width, poltype, fwhm):
    if poltype == "qu":
        pol_a = sfp.hex_pol_angles_qu(npix)
        pol_b = sfp.hex_pol_angles_qu(npix, offset=90.0)
    elif poltype == "radial":
        pol_a = sfp.hex_pol_angles_radial(npix)
        pol_b = sfp.hex_pol_angles_radial(npix, offset=90.0)
    dets_a = sfp.hex_layout(npix, width, "", "A", pol_a)
    dets_b = sfp.hex_layout(npix, width, "", "B", pol_b)

    dets = dict()
    dets.update(dets_a)
    dets.update(dets_b)

    # Pol color different for A/B detectors
    detpolcolor = dict()
    detpolcolor.update({x: "red" for x in dets_a.keys()})
    detpolcolor.update({x: "blue" for x in dets_b.keys()})

    # set the label to just the detector name
    detlabels = {x: x for x in dets.keys()}

    # fwhm and face color the same
    detfwhm = {x: fwhm for x in dets.keys()}

    # cycle through some colors just for fun
    pclr = [
        (1.0, 0.0, 0.0, 0.1),
        (1.0, 0.5, 0.0, 0.1),
        (0.25, 0.5, 1.0, 0.1),
        (0.0, 0.75, 0.0, 0.1),
    ]
    detcolor = {y: pclr[(x // 2) % 4] for x, y in enumerate(sorted(dets.keys()))}

    # split out quaternions for plotting
    detquats = {x: dets[x]["quat"] for x in dets.keys()}

    return dets, detquats, detfwhm, detcolor, detpolcolor, detlabels


def generate_rhombus(npix, width, fwhm, prefix, center):
    pol_a = sfp.rhomb_pol_angles_qu(npix)
    pol_b = sfp.rhomb_pol_angles_qu(npix, offset=90.0)
    dets_a = sfp.rhombus_layout(npix, width, prefix, "A", pol_a, center=center)
    dets_b = sfp.rhombus_layout(npix, width, prefix, "B", pol_b, center=center)

    dets = dict()
    dets.update(dets_a)
    dets.update(dets_b)

    # Pol color different for A/B detectors
    detpolcolor = dict()
    detpolcolor.update({x: "red" for x in dets_a.keys()})
    detpolcolor.update({x: "blue" for x in dets_b.keys()})

    # set the label to just the detector name
    detlabels = {x: x for x in dets.keys()}

    # fwhm and face color the same
    detfwhm = {x: fwhm for x in dets.keys()}

    # cycle through some colors just for fun
    pclr = [
        (1.0, 0.0, 0.0, 0.1),
        (1.0, 0.5, 0.0, 0.1),
        (0.25, 0.5, 1.0, 0.1),
        (0.0, 0.75, 0.0, 0.1),
    ]
    detcolor = {y: pclr[(x // 2) % 4] for x, y in enumerate(sorted(dets.keys()))}

    # split out quaternions for plotting
    detquats = {x: dets[x]["quat"] for x in dets.keys()}

    return dets, detquats, detfwhm, detcolor, detpolcolor, detlabels


class SimFocalplaneTest(MPITestCase):
    def setUp(self):
        self.outdir = "toast_test_output"
        self.rank = 0
        if self.comm is not None:
            self.rank = self.comm.rank
        if self.rank == 0:
            if not os.path.isdir(self.outdir):
                os.mkdir(self.outdir)

    def test_cart_quat(self):
        xincr = np.linspace(-5.0, 5.0, num=10, endpoint=True)
        yincr = np.linspace(-5.0, 5.0, num=10, endpoint=True)
        offsets = list()
        for x in xincr:
            for y in yincr:
                ang = 3.6 * (x - xincr[0]) * (y - yincr[0])
                offsets.append([x, y, ang])
        quats = sfp.cartesian_to_quat(offsets)
        detquats = {"{}".format(x): y for x, y in enumerate(quats)}
        fwhm = {x: 30.0 for x in detquats.keys()}
        outfile = os.path.join(self.outdir, "out_test_cart2quat.png")
        if self.rank == 0:
            sfp.plot_focalplane(detquats, 12.0, 12.0, outfile, fwhm=fwhm)
        return

    def test_hex_nring(self):
        result = {
            1: 1,
            7: 2,
            19: 3,
            37: 4,
            61: 5,
            91: 6,
            127: 7,
            169: 8,
            217: 9,
            271: 10,
            331: 11,
            397: 12,
        }
        for npix, check in result.items():
            test = sfp.hex_nring(npix)
            nt.assert_equal(test, check)
        return

    def test_vis_hex_small(self):
        dets, detquats, detfwhm, detcolor, detpolcolor, detlabels = generate_hex(
            7, 5.0, "qu", 15.0
        )
        outfile = os.path.join(self.outdir, "out_test_vis_hex_small.png")
        if self.rank == 0:
            sfp.plot_focalplane(
                detquats,
                6.0,
                6.0,
                outfile,
                fwhm=detfwhm,
                facecolor=detcolor,
                polcolor=detpolcolor,
                labels=detlabels,
            )
        return

    def test_vis_hex_small_rad(self):
        dets, detquats, detfwhm, detcolor, detpolcolor, detlabels = generate_hex(
            7, 5.0, "radial", 15.0
        )
        outfile = os.path.join(self.outdir, "out_test_vis_hex_small_rad.png")
        if self.rank == 0:
            sfp.plot_focalplane(
                detquats,
                6.0,
                6.0,
                outfile,
                fwhm=detfwhm,
                facecolor=detcolor,
                polcolor=detpolcolor,
                labels=detlabels,
            )
        return

    def test_vis_hex_medium(self):
        dets, detquats, detfwhm, detcolor, detpolcolor, detlabels = generate_hex(
            91, 5.0, "qu", 10.0
        )
        outfile = os.path.join(self.outdir, "out_test_vis_hex_medium.png")
        if self.rank == 0:
            sfp.plot_focalplane(
                detquats,
                6.0,
                6.0,
                outfile,
                fwhm=detfwhm,
                facecolor=detcolor,
                polcolor=detpolcolor,
                labels=detlabels,
            )
        return

    def test_vis_hex_large(self):
        dets, detquats, detfwhm, detcolor, detpolcolor, detlabels = generate_hex(
            217, 5.0, "qu", 5.0
        )
        outfile = os.path.join(self.outdir, "out_test_vis_hex_large.png")
        if self.rank == 0:
            sfp.plot_focalplane(
                detquats,
                6.0,
                6.0,
                outfile,
                fwhm=detfwhm,
                facecolor=detcolor,
                polcolor=detpolcolor,
                labels=detlabels,
            )
        return

    def test_vis_rhombus(self):
        sixty = np.pi / 3.0
        thirty = np.pi / 6.0
        rtthree = np.sqrt(3.0)

        rdim = 8
        rpix = rdim ** 2

        hexwidth = 5.0
        rwidth = hexwidth / rtthree

        # angular separation of rhombi
        margin = 0.60 * hexwidth

        centers = [
            np.array([0.5 * margin, 0.0, 0.0]),
            np.array(
                [-0.5 * np.cos(sixty) * margin, 0.5 * np.sin(sixty) * margin, 120.0]
            ),
            np.array(
                [-0.5 * np.cos(sixty) * margin, -0.5 * np.sin(sixty) * margin, 240.0]
            ),
        ]

        cquats = sfp.cartesian_to_quat(centers)

        dets = dict()
        detquats = dict()
        detfwhm = dict()
        detcolor = dict()
        detpolcolor = dict()
        detlabels = dict()
        for w, c in enumerate(cquats):
            (
                wdets,
                wdetquats,
                wdetfwhm,
                wdetcolor,
                wdetpolcolor,
                wdetlabels,
            ) = generate_rhombus(rpix, rwidth, 7.0, "{}".format(w), c)
            dets.update(wdets)
            detquats.update(wdetquats)
            detfwhm.update(wdetfwhm)
            detcolor.update(wdetcolor)
            detpolcolor.update(wdetpolcolor)
            detlabels.update(wdetlabels)

        outfile = os.path.join(self.outdir, "out_test_vis_rhombus.png")
        if self.rank == 0:
            sfp.plot_focalplane(
                detquats,
                1.2 * hexwidth,
                1.2 * hexwidth,
                outfile,
                fwhm=detfwhm,
                facecolor=detcolor,
                polcolor=detpolcolor,
                labels=detlabels,
            )
        return
